# Copyright 2022 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from http import HTTPStatus

import pytest

from test.unit_tests.fixtures.workflows import add_soft_deleted_workflows  # noqa: F401


@pytest.fixture(autouse=True)
def _add_soft_deleted_workflows(add_soft_deleted_workflows):  # noqa: F811
    add_soft_deleted_workflows(10)


process_fields = [
    "processId",
    "isTask",
    "lastStep",
    "lastStatus",
    "assignee",
    "failedReason",
    "traceback",
    "workflowId",
    "workflowName",
    "createdBy",
    "startedAt",
    "lastModifiedAt",
    "product",
]


def get_processes_query(
    first: int = 10,
    after: int = 0,
    filter_by: list[dict[str, str]] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    query_string: str | None = None,
) -> bytes:
    query = """
query ProcessQuery($first: Int!, $after: Int!, $sortBy: [GraphqlSort!], $filterBy: [GraphqlFilter!], $query: String) {
  processes(first: $first, after: $after, sortBy: $sortBy, filterBy: $filterBy, query: $query) {
    page {
      processId
      isTask
      lastStep
      lastStatus
      assignee
      failedReason
      traceback
      workflowId
      workflowName
      createdBy
      startedAt
      lastModifiedAt
      form
      product {
        productId
        name
        description
        productType
        status
        tag
        createdAt
        endDate
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
}
    """
    return json.dumps(
        {
            "operationName": "ProcessQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def get_processes_query_with_subscriptions(
    first: int = 10,
    after: int = 0,
    filter_by: list[str] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    query_string: str | None = None,
) -> bytes:
    query = """
query ProcessQuery($first: Int!, $after: Int!, $sortBy: [GraphqlSort!], $filterBy: [GraphqlFilter!], $query: String) {
  processes(first: $first, after: $after, sortBy: $sortBy, filterBy: $filterBy, query: $query) {
    page {
      processId
      isTask
      lastStep
      lastStatus
      assignee
      failedReason
      traceback
      workflowId
      workflowName
      createdBy
      startedAt
      lastModifiedAt
      product {
        productId
        name
        description
        productType
        status
        tag
        createdAt
        endDate
      }
      subscriptions {
        page {
          description
          subscriptionId
          status
          insync
          note
          startDate
          endDate
        }
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
}
    """
    return json.dumps(
        {
            "operationName": "ProcessQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def get_processes_state_updates_and_delta(
    first: int = 10,
    after: int = 0,
    filter_by: list[dict[str, str]] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    query_string: str | None = None,
) -> bytes:
    query = """
query ProcessQuery($first: Int!, $after: Int!, $sortBy: [GraphqlSort!], $filterBy: [GraphqlFilter!], $query: String) {
  processes(first: $first, after: $after, sortBy: $sortBy, filterBy: $filterBy, query: $query) {
    page {
      steps {
        stateDelta
        status
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
}
    """
    return json.dumps(
        {
            "operationName": "ProcessQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def test_processes_has_next_page(
    test_client,
    mocked_processes,
    mocked_processes_resumeall,
    generic_subscription_2,
    generic_subscription_1,
):
    data = get_processes_query()
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    assert len(processes) == 10

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": 19,
    }

    for process in processes:
        for field in process_fields:
            assert field in process


def test_process_has_previous_page(
    test_client,
    mocked_processes,
    mocked_processes_resumeall,
    generic_subscription_2,
    generic_subscription_1,
):
    data = get_processes_query(after=1)
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    assert len(processes) == 10

    assert pageinfo == {
        "hasPreviousPage": True,
        "hasNextPage": True,
        "startCursor": 1,
        "endCursor": 10,
        "totalItems": 19,
    }


def test_processes_sorting_asc(
    test_client,
    mocked_processes,
    mocked_processes_resumeall,
    generic_subscription_2,
    generic_subscription_1,
):
    # when

    data = get_processes_query(sort_by=[{"field": "startedAt", "order": "ASC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": 19,
    }

    assert processes[0]["startedAt"] == "2020-01-14T09:30:00+00:00"
    assert processes[1]["startedAt"] == "2020-01-15T09:30:00+00:00"
    assert processes[2]["startedAt"] == "2020-01-15T09:30:00+00:00"


def test_processes_sorting_desc(
    test_client,
    mocked_processes,
    mocked_processes_resumeall,
    generic_subscription_2,
    generic_subscription_1,
):
    # when

    data = get_processes_query(sort_by=[{"field": "startedAt", "order": "DESC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": 19,
    }

    assert processes[0]["startedAt"] == "2020-02-19T09:30:00+00:00"
    assert processes[1]["startedAt"] == "2020-02-19T09:30:00+00:00"
    assert processes[2]["startedAt"] == "2020-02-18T09:30:00+00:00"
    assert processes[3]["startedAt"] == "2020-02-17T09:30:00+00:00"
    assert processes[4]["startedAt"] == "2020-02-16T09:30:00+00:00"


@pytest.mark.parametrize(
    "query_args",
    [
        {"filter_by": [{"field": "lastStatus", "value": "completed"}]},
        {"query_string": "lastStatus:completed"},
        {"query_string": "last_status:completed"},
    ],
)
def test_processes_has_filtering(
    test_client,
    mocked_processes,
    mocked_processes_resumeall,
    generic_subscription_2,
    generic_subscription_1,
    query_args,
):
    # when

    data = get_processes_query(**query_args)
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    # then

    assert HTTPStatus.OK == response.status_code

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 3,
        "totalItems": 4,
    }

    for process in processes:
        assert process["lastStatus"] == "COMPLETED"


def test_processes_filtering_with_invalid_filter(
    test_client,
    mocked_processes,
    mocked_processes_resumeall,
    generic_subscription_2,
    generic_subscription_1,
):
    # when

    data = get_processes_query(
        filter_by=[{"field": "lastStatus", "value": "completed"}, {"field": "test", "value": "invalid"}]
    )
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    errors = result["errors"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    assert errors == [
        {
            "message": (
                "Invalid filter arguments (invalid_filters=['test'] "
                "valid_filter_keys=['assignee', 'createdBy', "
                "'customer', 'failedReason', 'isTask', 'lastModifiedAt', 'lastStatus', "
                "'lastStep', 'processId', 'product', 'productDescription', 'startedAt', "
                "'subscriptionId', 'tag', 'target', 'traceback', 'workflowId', "
                "'workflowName'])"
            ),
            "path": ["processes"],
            "extensions": {"error_type": "internal_error"},
        }
    ]
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 3,
        "totalItems": 4,
    }

    for process in processes:
        assert process["lastStatus"] == "COMPLETED"


@pytest.mark.parametrize(
    "query_args,num_results",
    [
        ({"query_string": "isTask:no", "first": 20}, 0),
        ({"filter_by": [{"field": "isTask", "value": "false"}], "first": 20}, 6),
        ({"query_string": 'isTask:"true"', "first": 20}, 13),
        ({"query_string": "isTask:(true|false)", "first": 20}, 19),
        ({"query_string": "nonsense"}, 0),
        ({"filter_by": [{"field": "assignee", "value": ""}]}, 0),
        ({"query_string": None, "first": 100}, 19),
        ({"query_string": "one"}, 8),
        ({"query_string": "two"}, 7),
        ({"query_string": "product:(1 | 3)"}, 8),
        ({"query_string": "tag:gen1"}, 8),
        ({"query_string": "tag:gEN2"}, 7),
        ({"query_string": "-tag:gen*"}, 4),
    ],
)
def test_processes_various_filterings(
    test_client,
    mocked_processes,
    mocked_processes_resumeall,
    generic_subscription_2,
    generic_subscription_1,
    query_args,
    num_results,
):
    # when

    data = get_processes_query(**query_args)
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    assert len(processes) == num_results


@pytest.mark.parametrize(
    "query_args",
    [
        lambda pid: {"filter_by": [{"field": "processId", "value": pid}]},
        lambda pid: {"query_string": f"processId:{pid}"},
        lambda pid: {"query_string": f"process_id:{pid}"},
    ],
)
def test_single_process(
    test_client,
    mocked_processes,
    mocked_processes_resumeall,
    generic_subscription_2,
    generic_subscription_1,
    query_args,
):
    process_pid = str(mocked_processes[0])
    # when

    data = get_processes_query(**query_args(process_pid))
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    assert len(processes) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert processes[0]["processId"] == process_pid


def test_single_process_with_form(
    test_client,
    mocked_processes,
    generic_subscription_2,
    generic_subscription_1,
):
    process_pid = str(mocked_processes[1])
    # when

    data = get_processes_query(filter_by=[{"field": "process_id", "value": process_pid}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]
    assert len(processes) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert processes[0]["processId"] == process_pid
    assert processes[0]["form"] == {
        "title": "unknown",
        "type": "object",
        "properties": {"generic_select": {"$ref": "#/$defs/TestChoice"}},
        "additionalProperties": False,
        "required": ["generic_select"],
        "$defs": {"TestChoice": {"enum": ["A", "B", "C"], "title": "TestChoice", "type": "string"}},
    }


@pytest.mark.parametrize(
    "query_args",
    [
        lambda pid: {"filter_by": [{"field": "processId", "value": pid}]},
        lambda pid: {"query_string": f"processId:{pid}"},
        lambda pid: {"query_string": f"process_id:{pid}"},
    ],
)
def test_single_process_with_subscriptions(
    test_client,
    mocked_processes,
    mocked_processes_resumeall,
    generic_subscription_2,
    generic_subscription_1,
    query_args,
):
    process_pid = str(mocked_processes[0])
    # when

    data = get_processes_query_with_subscriptions(**query_args(process_pid))
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    assert len(processes) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert processes[0]["processId"] == process_pid
    assert processes[0]["subscriptions"]["page"][0]["subscriptionId"] == generic_subscription_1


def test_processes_sorting_product_tag_asc(
    test_client,
    mocked_processes,
    mocked_processes_resumeall,
    generic_subscription_2,
    generic_subscription_1,
):
    # when

    data = get_processes_query(sort_by=[{"field": "productTag", "order": "ASC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": 15,
    }

    assert [process["product"]["tag"] for process in processes] == [
        "GEN1",
        "GEN1",
        "GEN1",
        "GEN1",
        "GEN1",
        "GEN1",
        "GEN1",
        "GEN1",
        "GEN2",
        "GEN2",
    ]


@pytest.mark.parametrize(
    "query_args",
    [
        lambda pid: {"filter_by": [{"field": "processId", "value": pid}]},
        lambda pid: {"query_string": f"processId:{pid}"},
        lambda pid: {"query_string": f"process_id:{pid}"},
    ],
)
def test_processes_state_updates_and_delta(
    test_client,
    mocked_processes,
    mocked_processes_resumeall,
    generic_subscription_2,
    generic_subscription_1,
    query_args,
):
    # when

    process_pid = str(mocked_processes[0])
    data = get_processes_state_updates_and_delta(**query_args(process_pid))
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    assert "errors" not in result
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]

    assert [process["steps"] for process in processes] == [
        [
            {"stateDelta": {}, "status": "success"},
            {
                "stateDelta": {"subscription_id": generic_subscription_1},
                "status": "success",
            },
            {"stateDelta": {}, "status": "success"},
            {"stateDelta": {}, "status": "suspend"},
            {"stateDelta": None, "status": "pending"},
            {"stateDelta": None, "status": "pending"},
        ]
    ]



# Copyright 2022 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from http import HTTPStatus


def build_simple_query(process_id):
    q = """
        query ProcessQuery($processId: UUID!) {
            process(processId: $processId) {
                processId
            }
        }
        """
    return json.dumps(
        {
            "operationName": "ProcessQuery",
            "query": q,
            "variables": {
                "processId": str(process_id),
            },
        }
    ).encode("utf-8")


def test_process(test_client, mocked_processes):
    process_id = mocked_processes[0]
    test_query = build_simple_query(process_id)

    response = test_client.post("/api/graphql", content=test_query, headers={"Content-Type": "application/json"})
    assert response.status_code == HTTPStatus.OK
    assert response.json() == {"data": {"process": {"processId": str(process_id)}}}



from itertools import chain
from typing import Any, NamedTuple
from uuid import UUID

import structlog
from sqlalchemy import Row, select
from sqlalchemy import Text as SaText
from sqlalchemy import cast as sa_cast
from sqlalchemy.orm import aliased
from strawberry.dataloader import DataLoader

from orchestrator.db import (
    ResourceTypeTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionTable,
    db,
)
from orchestrator.db.models import (
    SubscriptionInstanceRelationTable,
)
from orchestrator.services.subscriptions import RELATION_RESOURCE_TYPES
from orchestrator.types import SubscriptionLifecycle

logger = structlog.get_logger(__name__)


class Relation(NamedTuple):
    depends_on_sub_id: UUID
    in_use_by_sub_id: UUID


def _get_instance_relations(instance_relations_query: Any) -> list[Relation]:
    def to_relation(row: Row[Any]) -> Relation:
        return Relation(row[0], row[1])

    return [to_relation(row) for row in db.session.execute(instance_relations_query)]


async def _get_in_use_by_instance_relations(subscription_ids: list[UUID], filter_statuses: list[str]) -> list[Relation]:
    """Get in_use_by by relations through subscription instance hierarchy."""
    in_use_by_subscriptions = aliased(SubscriptionTable)
    in_use_by_instances = aliased(SubscriptionInstanceTable)
    depends_on_instances = aliased(SubscriptionInstanceTable)

    query_get_in_use_by_ids = (
        select(depends_on_instances.subscription_id, in_use_by_instances.subscription_id)
        .distinct()
        .join(in_use_by_instances.subscription)
        .join(in_use_by_instances.depends_on_block_relations)
        .join(depends_on_instances, SubscriptionInstanceRelationTable.depends_on)
        .join(in_use_by_subscriptions, depends_on_instances.subscription)
        .filter(depends_on_instances.subscription_id.in_(set(subscription_ids)))
        .filter(in_use_by_instances.subscription_id != depends_on_instances.subscription_id)
        .filter(in_use_by_subscriptions.status.in_(filter_statuses))
    )

    return _get_instance_relations(query_get_in_use_by_ids)


async def _get_depends_on_instance_relations(
    subscription_ids: list[UUID], filter_statuses: list[str]
) -> list[Relation]:
    """Get depends_on relations through subscription instance hierarchy."""
    in_use_by_instances = aliased(SubscriptionInstanceTable)
    depends_on_instances = aliased(SubscriptionInstanceTable)
    depends_on_subscriptions = aliased(SubscriptionTable)

    query_get_depends_on_ids = (
        select(depends_on_instances.subscription_id, in_use_by_instances.subscription_id)
        .distinct()
        .join(depends_on_instances.subscription)
        .join(depends_on_instances.in_use_by_block_relations)
        .join(in_use_by_instances, SubscriptionInstanceRelationTable.in_use_by)
        .join(depends_on_subscriptions, in_use_by_instances.subscription)
        .filter(in_use_by_instances.subscription_id.in_(set(subscription_ids)))
        .filter(depends_on_instances.subscription_id != in_use_by_instances.subscription_id)
        .filter(depends_on_subscriptions.status.in_(filter_statuses))
    )

    return _get_instance_relations(query_get_depends_on_ids)


def _get_resource_type_relations(resource_type_relations_query: Any) -> list[Relation]:
    def to_relation(row: Row[Any]) -> Relation:
        return Relation(UUID(row[0]), row[1])

    return [to_relation(row) for row in db.session.execute(resource_type_relations_query)]


async def _get_in_use_by_resource_type_relations(
    subscription_ids: list[UUID], filter_statuses: list[str]
) -> list[Relation]:
    """Get in_use_by relations through resource types."""
    logger.warning("Using legacy RELATION_RESOURCE_TYPES to find in_use_by subs")

    in_use_by_subscriptions = aliased(SubscriptionTable)
    depends_on_instance_values = aliased(SubscriptionInstanceValueTable)

    # Convert UUIDs to string
    unique_subscription_ids = set(map(str, subscription_ids))

    query_get_in_use_by_ids = (
        select(depends_on_instance_values.value, in_use_by_subscriptions.subscription_id)
        .select_from(depends_on_instance_values)
        .join(SubscriptionInstanceTable)
        .join(in_use_by_subscriptions)
        .join(ResourceTypeTable)
        .filter(ResourceTypeTable.resource_type.in_(RELATION_RESOURCE_TYPES))
        .filter(depends_on_instance_values.value.in_(unique_subscription_ids))
        .filter(in_use_by_subscriptions.status.in_(filter_statuses))
    )

    return _get_resource_type_relations(query_get_in_use_by_ids)


async def _get_depends_on_resource_type_relations(
    subscription_ids: list[UUID], filter_statuses: list[str]
) -> list[Relation]:
    """Get depends_on relations through resource types."""
    logger.warning("Using legacy RELATION_RESOURCE_TYPES to find depends_on subs")

    depends_on_subscriptions = aliased(SubscriptionTable)
    in_use_by_instances = aliased(SubscriptionInstanceTable)
    in_use_by_instance_values = aliased(SubscriptionInstanceValueTable)

    unique_subscription_ids = set(subscription_ids)

    query_get_depends_on_ids = (
        select(in_use_by_instance_values.value, in_use_by_instances.subscription_id)
        .select_from(in_use_by_instance_values)
        .join(in_use_by_instances)
        .join(
            depends_on_subscriptions,
            in_use_by_instance_values.value == sa_cast(depends_on_subscriptions.subscription_id, SaText),
        )
        .join(ResourceTypeTable)
        .filter(ResourceTypeTable.resource_type.in_(RELATION_RESOURCE_TYPES))
        .filter(in_use_by_instances.subscription_id.in_(unique_subscription_ids))
        .filter(depends_on_subscriptions.status.in_(filter_statuses))
    )

    return _get_resource_type_relations(query_get_depends_on_ids)


async def _get_in_use_by_relations(subscription_ids: list[UUID], filter_statuses: list[str]) -> list[Relation]:
    if RELATION_RESOURCE_TYPES:
        # Find relations through resource types
        resource_type_relations = await _get_in_use_by_resource_type_relations(subscription_ids, filter_statuses)
    else:
        resource_type_relations = []
    # Find relations through instance hierarchy
    instance_relations = await _get_in_use_by_instance_relations(subscription_ids, filter_statuses)
    return list(chain(resource_type_relations, instance_relations))


async def _get_depends_on_relations(subscription_ids: list[UUID], filter_statuses: list[str]) -> list[Relation]:
    if RELATION_RESOURCE_TYPES:
        # Find relations through resource types
        resource_type_relations = await _get_depends_on_resource_type_relations(subscription_ids, filter_statuses)
    else:
        resource_type_relations = []
    # Find relations through instance hierarchy
    instance_relations = await _get_depends_on_instance_relations(subscription_ids, filter_statuses)
    return list(chain(resource_type_relations, instance_relations))


async def in_use_by_subs_loader(keys: list[tuple[UUID, list[str] | None]]) -> list[list[SubscriptionTable]]:
    """GraphQL dataloader to efficiently get the in_use_by SubscriptionTables for multiple subscription_ids."""
    subscription_ids = [key[0] for key in keys]
    filter_statuses: list[str] = keys[0][1] or SubscriptionLifecycle.values()

    in_use_by_relations = await _get_in_use_by_relations(subscription_ids, filter_statuses)

    # Retrieve SubscriptionTable for all unique inuseby ids
    unique_in_use_by_ids = {row.in_use_by_sub_id for row in in_use_by_relations}
    _in_use_by_subs = db.session.execute(
        select(SubscriptionTable).filter(SubscriptionTable.subscription_id.in_(unique_in_use_by_ids))
    ).scalars()
    in_use_by_subs = {subscription.subscription_id: subscription for subscription in _in_use_by_subs}

    # group (more_itertools.bucket doesn't seem to work for tuple of uuids)
    subscription_in_use_by_ids: dict[UUID, list[UUID]] = {}
    for relation in in_use_by_relations:
        subscription_in_use_by_ids.setdefault(relation.depends_on_sub_id, []).append(relation.in_use_by_sub_id)

    def get_in_use_by_subs(depends_on_id: UUID) -> list[SubscriptionTable]:
        in_use_by_ids = subscription_in_use_by_ids.get(depends_on_id, [])
        return [in_use_by_sub for id_ in in_use_by_ids if (in_use_by_sub := in_use_by_subs.get(id_))]

    # Important (as with any dataloader)
    # Return the list of inuseby subs in the exact same order as the ids passed to this function
    return [get_in_use_by_subs(subscription_id) for subscription_id in subscription_ids]


async def depends_on_subs_loader(keys: list[tuple[UUID, list[str] | None]]) -> list[list[SubscriptionTable]]:
    """GraphQL dataloader to efficiently get the depends_on SubscriptionTables for multiple subscription_ids."""
    subscription_ids = [key[0] for key in keys]
    filter_statuses: list[str] = keys[0][1] or SubscriptionLifecycle.values()

    depends_on_relations = await _get_depends_on_relations(subscription_ids, filter_statuses)

    # Retrieve SubscriptionTable for all unique dependson ids
    unique_depends_on_ids = {row.depends_on_sub_id for row in depends_on_relations}
    _depends_on_subs = db.session.execute(
        select(SubscriptionTable).filter(SubscriptionTable.subscription_id.in_(unique_depends_on_ids))
    ).scalars()
    depends_on_subs = {subscription.subscription_id: subscription for subscription in _depends_on_subs}

    # group (more_itertools.bucket doesn't seem to work for tuple of uuids)
    subscription_depends_on_ids: dict[UUID, list[UUID]] = {}
    for relation in depends_on_relations:
        subscription_depends_on_ids.setdefault(relation.in_use_by_sub_id, []).append(relation.depends_on_sub_id)

    def get_depends_on_subs(in_use_by_id: UUID) -> list[SubscriptionTable]:
        depends_on_ids = subscription_depends_on_ids.get(in_use_by_id, [])
        return [depends_on_sub for id_ in depends_on_ids if (depends_on_sub := depends_on_subs.get(id_))]

    # Important (as with any dataloader)
    # Return the list of dependson subs in the exact same order as the ids passed to this function
    return [get_depends_on_subs(subscription_id) for subscription_id in subscription_ids]


SubsLoaderType = DataLoader[tuple[UUID, list[str] | None], list[SubscriptionTable]]



import json

from sqlalchemy import select, text

from orchestrator.cli.database import migrate_domain_models
from orchestrator.db import db
from orchestrator.db.models import ResourceTypeTable, SubscriptionInstanceValueTable
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import ProductBlockModel, SubscriptionModel
from orchestrator.services.resource_types import get_resource_types
from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.fixtures.products.product_blocks.product_block_one import DummyEnum


def test_migrate_domain_models_new_product_block(
    test_product_type_one, test_product_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, ProductBlockOneForTest = test_product_block_one

    class TestBlock(ProductBlockModel, product_block_name="test block", lifecycle=[SubscriptionLifecycle.ACTIVE]):
        str_field: str

    class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        new_block: TestBlock
        block: ProductBlockOneForTest

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    inputs = json.dumps(
        {
            "test block": {"description": "test block description", "tag": "test_block_tag"},
            "str_field": {"test block": "test"},
        }
    )
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs)

    def test_expected_before_upgrade():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert "new_block" not in subscription.block.model_dump()

    test_expected_before_upgrade()

    assert len(upgrade_sql) == 5
    assert len(downgrade_sql) == 5

    for stmt in upgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    new_model: ProductTypeOneForTestNew = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
    assert new_model.new_block
    assert new_model.new_block.str_field
    assert new_model.new_block.str_field == "test"

    for stmt in downgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    test_expected_before_upgrade()


def test_migrate_domain_models_new_product_block_on_product_block(
    test_product_type_one, test_product_block_one, product_one_subscription_1, test_product_sub_block_one
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, SubBlockOneForTest = test_product_sub_block_one

    class TestBlock(ProductBlockModel, product_block_name="test block", lifecycle=[SubscriptionLifecycle.ACTIVE]):
        str_field: str

    class ProductBlockOneForTestUpdated(
        ProductBlockModel, product_block_name="ProductBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        sub_block: SubBlockOneForTest
        sub_block_2: SubBlockOneForTest
        sub_block_list: list[SubBlockOneForTest]
        str_field: str
        int_field: int
        list_field: list[int]
        enum_field: DummyEnum
        new_block: TestBlock

    class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        block: ProductBlockOneForTestUpdated

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    inputs = json.dumps(
        {
            "test block": {"description": "test block description", "tag": "test_block_tag"},
            "str_field": {"test block": "test"},
        }
    )
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs)

    def test_expected_before_upgrade():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert "new_block" not in subscription.block.model_dump()

    test_expected_before_upgrade()

    assert len(upgrade_sql) == 6
    assert len(downgrade_sql) == 5

    for stmt in upgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    new_model: ProductTypeOneForTestNew = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
    assert new_model.block.new_block
    assert new_model.block.new_block.str_field
    assert new_model.block.new_block.str_field == "test"

    for stmt in downgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    test_expected_before_upgrade()


def test_migrate_domain_models_new_resource_type(
    test_product_type_one, test_product_sub_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, SubBlockOneForTest = test_product_sub_block_one

    class ProductBlockOneForTestUpdated(
        ProductBlockModel, product_block_name="ProductBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        sub_block: SubBlockOneForTest
        sub_block_2: SubBlockOneForTest
        sub_block_list: list[SubBlockOneForTest]
        int_field: int
        str_field: str
        list_field: list[int]
        enum_field: DummyEnum
        new_int_field: int

    class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        block: ProductBlockOneForTestUpdated

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    inputs = json.dumps({"new_int_field": {"ProductBlockOneForTest": "1", "description": "test new int field type"}})
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs)

    assert len(upgrade_sql) == 3
    assert len(downgrade_sql) == 4

    def test_expected_before_upgrade():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert "new_int_field" not in subscription.block.model_dump()

        new_int_field_resource = get_resource_types(filters=[ResourceTypeTable.resource_type == "new_int_field"])
        assert not new_int_field_resource

    test_expected_before_upgrade()

    for stmt in upgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    updated_subscription = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
    assert updated_subscription.block
    assert updated_subscription.block.new_int_field == 1

    for stmt in downgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    test_expected_before_upgrade()


def test_migrate_domain_models_new_product_block_and_resource_type(
    test_product_type_one, test_product_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, ProductBlockOneForTest = test_product_block_one

    class TestBlock(ProductBlockModel, product_block_name="test block", lifecycle=[SubscriptionLifecycle.ACTIVE]):
        str_field: str
        new_str_field: str

    class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        new_block: TestBlock
        block: ProductBlockOneForTest

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    inputs = json.dumps(
        {
            "test block": {"description": "test block description", "tag": "test_block_tag"},
            "str_field": {"test block": "test"},
            "new_str_field": {"description": "new str field desc", "test block": "new test"},
        }
    )
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs)

    def test_expected_before_upgrade():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert "new_block" not in subscription.block.model_dump()

    test_expected_before_upgrade()

    assert len(upgrade_sql) == 8
    assert len(downgrade_sql) == 9

    for stmt in upgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    new_model: ProductTypeOneForTestNew = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
    assert new_model.new_block
    assert new_model.new_block.str_field
    assert new_model.new_block.str_field == "test"
    assert new_model.new_block.new_str_field == "new test"

    for stmt in downgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    test_expected_before_upgrade()


def test_migrate_domain_models_update_resource_type(
    test_product_type_one, test_product_sub_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, SubBlockOneForTest = test_product_sub_block_one

    class ProductBlockOneForTestUpdated(
        ProductBlockModel, product_block_name="ProductBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        sub_block: SubBlockOneForTest
        sub_block_2: SubBlockOneForTest
        sub_block_list: list[SubBlockOneForTest]
        str_field: str
        int_field: int
        enum_field: DummyEnum
        new_list_field: list[int]

    class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        block: ProductBlockOneForTestUpdated

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    updates = json.dumps({"resource_types": {"list_field": "new_list_field"}})
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, updates=updates)

    assert len(upgrade_sql) == 1
    assert "UPDATE" in upgrade_sql[0]
    assert len(downgrade_sql) == 1
    assert "UPDATE" in downgrade_sql[0]

    def test_expected_before_upgrade():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert subscription.block.list_field == [10, 20, 30]
        assert "new_int_field" not in subscription.block.model_dump()

        new_int_field_resource = get_resource_types(filters=[ResourceTypeTable.resource_type == "new_int_field"])
        assert not new_int_field_resource

    test_expected_before_upgrade()

    for stmt in upgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    updated_subscription = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
    assert updated_subscription.block.new_list_field == [10, 20, 30]

    for stmt in downgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    test_expected_before_upgrade()


def test_migrate_domain_models_create_and_rename_resource_type(test_product_type_one, product_one_subscription_1):
    _, _, ProductTypeOneForTest = test_product_type_one

    class SubBlockOneForTestNewResource(
        ProductBlockModel, product_block_name="SubBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        int_field: int
        str_field: str
        new_list_field: list[int]

    class ProductBlockOneForTestUpdated(
        ProductBlockModel, product_block_name="ProductBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        sub_block: SubBlockOneForTestNewResource
        sub_block_2: SubBlockOneForTestNewResource
        sub_block_list: list[SubBlockOneForTestNewResource]
        str_field: str
        int_field: int
        enum_field: DummyEnum
        new_list_field: list[int]

    class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        block: ProductBlockOneForTestUpdated

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    inputs = json.dumps({"new_list_field": {"SubBlockOneForTest": "5"}})
    updates = json.dumps({"resource_types": {"list_field": "new_list_field"}})
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs=inputs, updates=updates)

    assert len(upgrade_sql) == 3
    assert [sql_stmt for sql_stmt in upgrade_sql if "UPDATE" in sql_stmt]
    assert len(downgrade_sql) == 3
    assert [sql_stmt for sql_stmt in downgrade_sql if "UPDATE" in sql_stmt]

    def test_expected_before_upgrade():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert subscription.block.list_field == [10, 20, 30]
        assert "new_int_field" not in subscription.block.model_dump()
        assert "new_int_field" not in subscription.block.sub_block.model_dump()

        new_int_field_resource = get_resource_types(filters=[ResourceTypeTable.resource_type == "new_int_field"])
        assert not new_int_field_resource

    test_expected_before_upgrade()

    for stmt in upgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    updated_instance = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
    assert updated_instance.block.new_list_field == [10, 20, 30]
    assert updated_instance.block.sub_block.new_list_field == [5]

    for stmt in downgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    test_expected_before_upgrade()


def test_migrate_domain_models_create_and_rename_and_delete_resource_type(
    test_product_type_sub_list_union,
    product_sub_list_union_subscription_1,
):
    _, _, ProductSubListUnion = test_product_type_sub_list_union

    class SubBlockOneForTestChanged(
        ProductBlockModel, product_block_name="SubBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        changed_int_field: int

    class SubBlockTwoForTestChanged(
        ProductBlockModel, product_block_name="SubBlockTwoForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        int_field_2: int
        changed_int_field: int

    class ProductBlockWithListUnionForTestNew(
        ProductBlockModel,
        product_block_name="ProductBlockWithListUnionForTest",
        lifecycle=[SubscriptionLifecycle.ACTIVE],
    ):
        list_union_blocks: list[SubBlockTwoForTestChanged | SubBlockOneForTestChanged]
        changed_int_field: int
        str_field: str
        list_field: list[int]

    class ProductSubListUnionTest(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_block: ProductBlockWithListUnionForTestNew

    SUBSCRIPTION_MODEL_REGISTRY["ProductSubListUnion"] = ProductSubListUnionTest

    inputs = json.dumps({"changed_int_field": {"SubBlockOneForTest": "5", "SubBlockTwoForTest": "2"}})
    updates = json.dumps({"resource_types": {"int_field": "changed_int_field"}})
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs=inputs, updates=updates)

    assert len(upgrade_sql) == 5
    assert [sql_stmt for sql_stmt in upgrade_sql if "UPDATE" in sql_stmt]
    assert len(downgrade_sql) == 3
    assert [sql_stmt for sql_stmt in downgrade_sql if "UPDATE" in sql_stmt]

    def test_expected_before_upgrade():
        subscription = ProductSubListUnion.from_subscription(product_sub_list_union_subscription_1)
        assert subscription.test_block.int_field == 1
        assert subscription.test_block.list_union_blocks[1].int_field == 1
        assert "int_field" not in subscription.test_block.list_union_blocks[0].model_dump()
        assert "changed_int_field" not in subscription.test_block.model_dump()
        assert "changed_int_field" not in subscription.test_block.list_union_blocks[0].model_dump()
        assert "changed_int_field" not in subscription.test_block.list_union_blocks[1].model_dump()

        new_int_field_resource = get_resource_types(filters=[ResourceTypeTable.resource_type == "new_int_field"])
        assert not new_int_field_resource

    test_expected_before_upgrade()

    for stmt in upgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    updated_subscription = ProductSubListUnionTest.from_subscription(product_sub_list_union_subscription_1)
    assert updated_subscription.test_block.changed_int_field == 1
    assert updated_subscription.test_block.list_union_blocks[0].changed_int_field == 2
    assert updated_subscription.test_block.list_union_blocks[1].changed_int_field == 1
    assert "int_field" not in updated_subscription.test_block.model_dump()
    assert "int_field" not in updated_subscription.test_block.list_union_blocks[0].model_dump()
    assert "int_field" not in updated_subscription.test_block.list_union_blocks[1].model_dump()

    int_field_resource = get_resource_types(filters=[ResourceTypeTable.resource_type == "int_field"])
    assert not int_field_resource

    for stmt in downgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    test_expected_before_upgrade()


def test_migrate_domain_models_remove_product(test_product_type_one, product_one_subscription_1):
    _, _, ProductTypeOneForTest = test_product_type_one

    del SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"]

    inputs = json.dumps({"force_delete_products": "no"})
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs)

    assert len(upgrade_sql) == 7
    assert len(downgrade_sql) == 0

    for stmt in upgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTest


def test_migrate_domain_models_remove_fixed_input(
    test_product_type_one, test_product_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, ProductBlockOneForTest = test_product_block_one

    class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        block: ProductBlockOneForTest

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    upgrade_sql, downgrade_sql = migrate_domain_models("example", True)

    assert len(upgrade_sql) == 1
    assert len(downgrade_sql) == 1

    def test_expected_before_upgrade():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert not subscription.test_fixed_input

    test_expected_before_upgrade()

    for stmt in upgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    subscription = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
    assert "test_fixed_input" not in subscription.model_dump()

    for stmt in downgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    test_expected_before_upgrade()


def test_migrate_domain_models_remove_product_block(test_product_type_one, product_one_subscription_1):
    _, _, ProductTypeOneForTest = test_product_type_one

    class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    upgrade_sql, downgrade_sql = migrate_domain_models("example", True)

    assert len(upgrade_sql) == 3
    assert len(downgrade_sql) == 0

    subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
    assert subscription.block
    assert subscription.block.int_field

    int_field_resource: list[ResourceTypeTable] = get_resource_types(
        filters=[ResourceTypeTable.resource_type == "int_field"]
    )
    int_field_instance_values = db.session.scalars(
        select(SubscriptionInstanceValueTable).where(
            SubscriptionInstanceValueTable.resource_type_id == int_field_resource[0].resource_type_id
        )
    ).all()
    assert len(int_field_instance_values) == 3

    for stmt in upgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    subscription = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
    assert "block" not in subscription.model_dump()

    int_field_instance_values = db.session.scalars(
        select(SubscriptionInstanceValueTable).where(
            SubscriptionInstanceValueTable.resource_type_id == int_field_resource[0].resource_type_id
        )
    ).all()
    assert len(int_field_instance_values) == 0

    for stmt in downgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()


def test_migrate_domain_models_remove_resource_type(
    test_product_type_one, test_product_sub_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, SubBlockOneForTest = test_product_sub_block_one

    class ProductBlockOneForTest(
        ProductBlockModel, product_block_name="ProductBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        sub_block: SubBlockOneForTest
        sub_block_2: SubBlockOneForTest
        sub_block_list: list[SubBlockOneForTest]
        int_field: int
        str_field: str
        enum_field: DummyEnum

    class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        block: ProductBlockOneForTest

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    upgrade_sql, downgrade_sql = migrate_domain_models("example", True)

    assert len(upgrade_sql) == 4
    assert len(downgrade_sql) == 0

    subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
    assert subscription.block.list_field

    for stmt in upgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    subscription = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
    assert "list_field" not in subscription.block.model_dump()

    for stmt in downgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()


def test_migrate_domain_models_update_block_resource_type(
    test_product_type_one, test_product_sub_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one
    _, _, SubBlockOneForTest = test_product_sub_block_one

    class ProductBlockOneForTestUpdated(
        ProductBlockModel, product_block_name="ProductBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        sub_block: SubBlockOneForTest
        sub_block_2: SubBlockOneForTest
        sub_block_list: list[SubBlockOneForTest]
        update_str_field: str
        int_field: int
        list_field: list[int]
        enum_field: DummyEnum

    class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        block: ProductBlockOneForTestUpdated

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    inputs = json.dumps(
        {
            "update_str_field": {"description": "test"},
            "updates": {"block_resource_types": {"ProductBlockOneForTest": {"str_field": "update_str_field"}}},
        }
    )
    updates = json.dumps({"block_resource_types": {"ProductBlockOneForTest": {"str_field": "update_str_field"}}})
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs=inputs, updates=updates)

    def test_expected_before_upgrade():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert subscription.block.str_field
        assert "update_str_field" not in subscription.block.model_dump()

        new_str_field_resource = get_resource_types(filters=[ResourceTypeTable.resource_type == "update_str_field"])
        assert not new_str_field_resource
        return subscription.block.str_field

    assert len(upgrade_sql) == 3
    assert len(downgrade_sql) == 4

    str_field_value = test_expected_before_upgrade()

    for stmt in upgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    subscription = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
    assert subscription.block.update_str_field == str_field_value

    for stmt in downgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    test_expected_before_upgrade()


def test_migrate_domain_models_rename_and_update_block_resource_type(test_product_type_one, product_one_subscription_1):
    _, _, ProductTypeOneForTest = test_product_type_one

    class SubBlockOneForTesUpdated(
        ProductBlockModel, product_block_name="SubBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        int_field: int
        other_str_field: str

    class ProductBlockOneForTestUpdated(
        ProductBlockModel, product_block_name="ProductBlockOneForTest", lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        sub_block: SubBlockOneForTesUpdated
        sub_block_2: SubBlockOneForTesUpdated
        sub_block_list: list[SubBlockOneForTesUpdated]
        update_str_field: str
        int_field: int
        list_field: list[int]
        enum_field: DummyEnum

    class ProductTypeOneForTestNew(SubscriptionModel, is_base=True, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        block: ProductBlockOneForTestUpdated

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    inputs = json.dumps({"update_str_field": {"description": "test"}})
    updates = json.dumps(
        {
            "resource_types": {"str_field": "other_str_field"},
            "block_resource_types": {"ProductBlockOneForTest": {"other_str_field": "update_str_field"}},
        }
    )
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs=inputs, updates=updates)

    def test_expected_before_upgrade():
        subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
        assert subscription.block.str_field
        assert "update_str_field" not in subscription.block.model_dump()

        new_str_field_resource = get_resource_types(filters=[ResourceTypeTable.resource_type == "update_str_field"])
        assert not new_str_field_resource
        return subscription.block.str_field

    assert len(upgrade_sql) == 6
    assert len(downgrade_sql) == 5

    str_field_value = test_expected_before_upgrade()

    for stmt in upgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    subscription = ProductTypeOneForTestNew.from_subscription(product_one_subscription_1)
    assert subscription.block.update_str_field == str_field_value

    for stmt in downgrade_sql:
        db.session.execute(text(stmt))
    db.session.commit()

    test_expected_before_upgrade()



from typing import Any, cast

import strawberry
from strawberry.experimental.pydantic.conversion import _convert_from_pydantic_to_strawberry_type
from strawberry.field import StrawberryField

# Vendored convert_pydantic_model_to_strawberry_class from
# https://github.com/strawberry-graphql/strawberry/blob/d721eb33176cfe22be5e47f5bf2c21a4a022a6d6/strawberry/experimental/pydantic/conversion.py
# And patched it so data loading happens inside the `if field init:`
# TODO: when confirmed to work without side-effects/regressions, create Issue/PR in strawberry-graphql


def convert_pydantic_model_to_strawberry_class__patched(
    cls: Any,  # noqa: ANN001
    *,
    model_instance: Any = None,  # noqa: ANN001
    extra: Any = None,  # noqa: ANN001
) -> Any:
    extra = extra or {}
    kwargs = {}

    for field_ in cls.__strawberry_definition__.fields:
        field = cast(StrawberryField, field_)
        python_name = field.python_name

        # only convert and add fields to kwargs if they are present in the `__init__`
        # method of the class
        if field.init:
            data_from_extra = extra.get(python_name, None)
            data_from_model = getattr(model_instance, python_name, None) if model_instance else None

            kwargs[python_name] = _convert_from_pydantic_to_strawberry_type(
                field.type, data_from_model, extra=data_from_extra
            )

    return cls(**kwargs)


strawberry.experimental.pydantic.object_type.convert_pydantic_model_to_strawberry_class = (  # type: ignore
    convert_pydantic_model_to_strawberry_class__patched
)



from ipaddress import IPv4Address, IPv6Address
from uuid import uuid4

import pytest

from test.acceptance_tests.fixtures.test_orchestrator.devtools.populator.test_product_populator import (
    TestProductPopulator,
)


@pytest.mark.acceptance
def test_test_product(new_test_product):
    populator = TestProductPopulator(
        an_int=1,
        a_str="string",
        a_bool=False,
        an_uuid=str(uuid4()),
        an_ipv4=IPv4Address("10.0.0.1"),
        an_ipv6=IPv6Address("::cafe:babe:feed:face:dead:beef"),
    )

    populator.start_create_workflow()
    populator.run()



"""Helper script to rewrite your orchestrator codebase for orchestrator-core 2.0.0.

Refer to the 2.0 migration guide documentation for background.
"""

import re
import sys
from pathlib import Path
from subprocess import run
from typing import Iterable


def remove_imports(text: str, module: str, symbol: str) -> tuple[str, bool]:
    """Find imports and remove them.

    Assumes code is formatted through Black to keep the regex somewhat readable.
    """
    text_orig = text

    # single import from module (may have a #comment) -> remove line
    rgx = r"(from %s import \b%s\b(\s*#[^\n]*)*\n)" % (re.escape(module), symbol)
    text = re.sub(rgx, "", text)

    # middle or last of multiple imports from module -> strip symbol
    rgx = r"(from %s import .+)(, \b%s\b)" % (re.escape(module), symbol)
    text = re.sub(rgx, r"\1", text)

    # first of multiple imports from same module -> strip symbol
    rgx = r"(from %s import )\b%s\b, " % (re.escape(module), symbol)
    text = re.sub(rgx, r"\1", text)

    # multiline import -> remove line with symbol
    rgx_verbose = r"""(?P<before>^from\s%s\simport\s*\([^\n]*\n(?:^[^\n]+,\n)*)
                      (^\s*\b%s\b,[^\n]*\n)
                      (?P<after>(?:^[^\n]+,\n)*\)[^\n]*$)"""
    text = re.sub(rgx_verbose % (re.escape(module), symbol), r"\g<before>\g<after>", text, flags=re.M | re.X)
    return text, text_orig != text


def insert_import(text: str, import_stmt: str) -> str:
    # Find the first import line and add our line above that
    # Rely on ruff & black for formatting
    return re.sub(r"(^(?:from .+|import .+)$)", f"{import_stmt}\n" + r"\1", text, count=1, flags=re.M)


def find_and_remove_aliases(text: str, symbol: str) -> tuple[str, list[str]]:
    """In the given text find aliases of the given symbol and remove them.

    Return updated text and aliases removed.
    """
    rgx = r"(\b%s as (\w+))" % (symbol,)
    aliases = [aliasgroup for fullgroup, aliasgroup in re.findall(rgx, text)]
    newtext = re.sub(rgx, symbol, text)
    return newtext, aliases


def replace_words(text: str, words: list[str], replace: str) -> str:
    rgx = r"\b(%s)\b" % ("|".join(words),)
    return re.sub(rgx, replace, text)


def has_word(text: str, word: str) -> bool:
    return bool(re.search(r"\b%s\b" % (word,), text))


def rewrite_subscription_instance_lists(f: Path) -> list[str]:
    """Rewrite all SubscriptionInstanceList occurrences in a file and fix the imports.

    Refer to the "SubscriptionInstanceList" section in the 2.0 migration guide for details.

    Returns:
        names of any replaced SI Lists.
    """
    text = text_orig = f.read_text()
    sil = "SubscriptionInstanceList"
    if not has_word(text, sil):
        return []

    # Remove aliases in imports and in usages
    text, aliases = find_and_remove_aliases(text, sil)
    text = replace_words(text, aliases, sil) if aliases else text

    # Remove imports
    text, _changed = remove_imports(text, "orchestrator.domain.base", sil)

    # Replace SubscriptionInstanceList classes with types
    rgx = (
        r"^\s*class (?P<name>\w+)\([^)]*SubscriptionInstanceList\[(?P<type>(?:\w+|Union\[[\s\w,]+\]))\][^)]*\):\n"
        r"(?:^\s+min_items = (?P<min_items>(?:\w+|\d+))\n)*"
        r"(?:^\s+max_items = (?P<max_items>(?:\w+|\d+))\n)*"
    )

    def replace_si_list(match: re.Match) -> str:
        subscript_type = match.group("type")
        result = "\n%s = Annotated[list[%s], Len(%s)]\n"

        def len_params() -> Iterable[str]:
            if min_items := match.group("min_items"):
                yield f"min_length={min_items}"
            if max_items := match.group("max_items"):
                yield f"max_length={max_items}"

        return result % (match.group("name"), subscript_type, ", ".join(len_params()))

    names = [match[0] for match in re.findall(rgx, text, flags=re.M)]
    text = re.sub(rgx, replace_si_list, text, flags=re.M)

    symbols = {
        "Len": "from annotated_types import Len",
        "Annotated": "from typing import Annotated",
    }

    for symbol, import_stmt in symbols.items():
        if has_word(text_orig, symbol):
            # Assumes that if the symbol is already in the file, it was also imported
            continue
        if not has_word(text, symbol):
            # Not every symbol may be necessary
            continue
        text = insert_import(text, import_stmt)

    with f.open(mode="w"):
        f.write_text(text)
    return names


def move_import(f: Path, symbol: str, old_module: str, new_module: str) -> bool:
    text = f.read_text()
    text, changed = remove_imports(text, old_module, symbol)
    if not changed:
        return False
    text = insert_import(text, f"from {new_module} import {symbol}")
    with f.open(mode="w"):
        f.write_text(text)
    return True


re_serializable_property = re.compile(r"^(\s+)(@serializable_property)([^\n]*)\n", flags=re.MULTILINE)


def replace_serializable_props(f: Path) -> bool:
    """Replace @serializable_property with pydantic's @computed_field and updates imports.

    As an example, this changes:

      @serializable_property
      def title():
        ...

    To:

      @computed_field  # type: ignore[misc]
      @property
      def title():
        ...

    The type:ignore is recommended by pydantic to silence mypy; if you don't use mypy it's not needed.
    """
    text = f.read_text()
    text, changed = remove_imports(text, "orchestrator.domain.base", "serializable_property")
    if not changed:
        return False
    text = insert_import(text, "from pydantic import computed_field")
    text = re_serializable_property.sub(r"\1@computed_field  # type: ignore[misc]\3\1@property\n", text)
    with f.open(mode="w"):
        f.write_text(text)
    return True


def migrate_file(f: Path) -> int:
    imports = {
        "SI": move_import(f, "SI", "orchestrator.domain.base", "orchestrator.types"),
        "VlanRanges": move_import(f, "VlanRanges", "orchestrator.utils.vlans", "nwastdlib.vlans"),
        "ReadOnlyField_forms": move_import(f, "ReadOnlyField", "pydantic_forms.core", "pydantic_forms.validators"),
        "ReadOnlyField_core": move_import(f, "ReadOnlyField", "orchestrator.forms", "pydantic_forms.validators"),
        "pydantic BaseSettings": move_import(f, "BaseSettings", "pydantic", "pydantic_settings"),
    }
    lines = []
    if replaced_lists := ", ".join(rewrite_subscription_instance_lists(f)):
        lines.append(f"replaced subscription instance lists [{replaced_lists}]")
    if replace_serializable_props(f):
        lines.append("replaced serializable properties")
    lines.extend([f"Moved {k} import" for k, v in imports.items() if v])

    if lines:
        formatted_lines = "\n".join(f" - {line}" for line in lines)
        print(f"Updated {f.name:50s}\n{formatted_lines}")

    return bool(lines)


def run_tool(*args: str) -> bool:
    cmd = " ".join(args)
    try:
        r = run(args, capture_output=True)  # noqa: S603
        if r.returncode == 0:
            return True
        print(f"{cmd} failed:", r.stdout, r.stderr)
    except FileNotFoundError:
        print(f"{cmd }failed: could not find executable in the current venv")
    return False


def migrate(target_dir: Path) -> bool:
    abs_path = str(target_dir.resolve())

    def run_tools() -> bool:
        return run_tool("ruff", "--fix", abs_path) and run_tool("black", "--quiet", abs_path)

    print(f"\n### Verifing files in {abs_path}... ", end="")
    if not run_tools():
        print("Failed to verify files, aborting migration. Please resolve the errors.")
        return False
    print("Ok")

    files_migrated = files_checked = 0
    print(f"\n### Migrating files in {abs_path}")
    try:
        for f in target_dir.glob("**/*.py"):
            if migrate_file(f):
                files_migrated += 1
            files_checked += 1
    except KeyboardInterrupt:
        print("Interrupted...")

    print(f"\n### Migrated {files_migrated}/{files_checked} files in {abs_path}")

    print(f"\n### Formatting files in {abs_path}")
    return run_tools()


if __name__ == "__main__":
    try:
        _target_dir = Path(sys.argv[1])
        assert _target_dir.is_dir()
    except Exception:
        print("Need a directory as parameter")
        sys.exit(1)

    sys.exit(0 if migrate(_target_dir) else 1)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from inspect import isgeneratorfunction
from typing import cast
from uuid import UUID

from more_itertools import first_true
from pydantic import field_validator
from sqlalchemy import select

from orchestrator.db import ProductTable, SubscriptionTable, db
from orchestrator.forms.validators import ProductId
from orchestrator.services import subscriptions
from orchestrator.settings import app_settings
from orchestrator.targets import Target
from orchestrator.types import State, SubscriptionLifecycle
from orchestrator.utils.redis import caching_models_enabled
from orchestrator.utils.state import form_inject_args
from orchestrator.workflow import StepList, Workflow, conditional, done, init, make_workflow, step
from orchestrator.workflows.steps import (
    cache_domain_models,
    refresh_subscription_search_index,
    resync,
    set_status,
    store_process_subscription,
    unsync,
    unsync_unchecked,
)
from pydantic_forms.core import FormPage
from pydantic_forms.types import FormGenerator, InputForm, InputStepFunc, StateInputStepFunc


def _generate_new_subscription_form(_workflow_target: str, workflow_name: str) -> InputForm:
    class NewProductPage(FormPage):
        product: ProductId

        @field_validator("product")
        @classmethod
        def product_validator(cls, product_id: UUID) -> UUID:
            """Run validator for initial_input_forms to check if the product exists and that this workflow is valid to run for this product."""
            product = db.session.get(ProductTable, product_id)
            if product is None:
                raise ValueError("Product not found")

            # Check if there is no reason to prevent this workflow
            current_workflow = product.create_subscription_workflow_key()

            if current_workflow != workflow_name:
                raise ValueError("This workflow is not valid for this product")

            return product_id

    return NewProductPage


def wrap_create_initial_input_form(initial_input_form: InputStepFunc | None) -> StateInputStepFunc | None:
    """Wrap initial input for create workflows.

    This is needed because the frontend expects all create workflows to start with a page that only contains the product.
    It also expects the second page to have some user visible inputs and the product *again*.
    """

    def create_initial_input_form_generator(state: State) -> FormGenerator:
        workflow_target: str = state["workflow_target"]
        workflow_name: str = state["workflow_name"]

        product_user_input = yield _generate_new_subscription_form(workflow_target, workflow_name)

        product = db.session.scalars(
            select(ProductTable).where(ProductTable.product_id == product_user_input.product)
        ).one()

        begin_state = {"product": product.product_id, "product_name": product.name}

        if initial_input_form is None:
            return begin_state

        form = form_inject_args(initial_input_form)({**state, **begin_state})

        if isgeneratorfunction(initial_input_form):
            user_input = yield from cast(FormGenerator, form)
        else:
            user_input_model = yield cast(InputForm, form)
            user_input = user_input_model.model_dump()

        return {**begin_state, **user_input}

    return create_initial_input_form_generator


# TODO these translations should stay in the frontend See also #549
TRANSLATIONS = {
    "subscription.not_in_sync": "Subscription already has a running process or task",
    "subscription.relations_not_in_sync": "This subscription can not be modified because some related subscriptions are not insync",
    "subscription.no_modify_invalid_status": "This subscription can not be modified because of the status it has",
    "subscription.no_modify_parent_subscription": "This subscription can not be modified as it is used in other subscriptions",
    "subscription.no_modify_subscription_in_use_by_others": "This subscription can not be modified as it is used in other subscriptions",
    "subscription.no_modify_auto_negotiation": "This workflow is not valid for this subscription",
}


def _generate_modify_form(workflow_target: str, workflow_name: str) -> InputForm:
    class ModifySubscriptionPage(FormPage):
        # We use UUID instead of SubscriptionId here because we don't want the allowed_status check and
        # we do our own validation here.
        subscription_id: UUID

        @field_validator("subscription_id")
        @classmethod
        def subscription_validator(cls, subscription_id: UUID) -> UUID:
            """Run validator for initial_input_forms to check if the subscription exists and that this workflow is valid to run for this subscription."""
            subscription = db.session.get(SubscriptionTable, subscription_id)
            if subscription is None:
                raise ValueError("Subscription not found")

            # Check if there is no reason to prevent this workflow
            workflows = subscriptions.subscription_workflows(subscription)
            current_workflow = first_true(
                workflows[workflow_target.lower()], None, lambda wf: wf["name"] == workflow_name
            )

            if not current_workflow:
                raise ValueError("This workflow is not valid for this subscription")

            if "reason" in current_workflow:
                message = TRANSLATIONS.get(current_workflow["reason"], current_workflow["reason"])
                raise ValueError(f"This workflow cannot be started: {message}")

            return subscription_id

    return ModifySubscriptionPage


def wrap_modify_initial_input_form(initial_input_form: InputStepFunc | None) -> StateInputStepFunc | None:
    """Wrap initial input for modify and terminate workflows.

    This is needed because the frontend expects all modify workflows to start with a page that only contains the
    subscription id. It also expects the second page to have some user visible inputs and the subscription id *again*.
    """

    def create_initial_input_form_generator(state: State) -> FormGenerator:
        workflow_target: str = state["workflow_target"]
        workflow_name: str = state["workflow_name"]

        user_input = yield _generate_modify_form(workflow_target, workflow_name)

        subscription = SubscriptionTable.query.get(user_input.subscription_id)

        begin_state = {
            "subscription_id": str(subscription.subscription_id),
            "product": str(subscription.product_id),
            "customer_id": subscription.customer_id,
        }

        if initial_input_form is None:
            return begin_state

        form = form_inject_args(initial_input_form)({**state, **begin_state})

        if isgeneratorfunction(initial_input_form):
            user_input = yield from cast(FormGenerator, form)
        else:
            user_input_model = yield cast(InputForm, form)
            user_input = user_input_model.model_dump()

        return {**begin_state, **user_input}

    return create_initial_input_form_generator


modify_initial_input_form_generator = None


validate_initial_input_form_generator = wrap_modify_initial_input_form(modify_initial_input_form_generator)

push_domain_models = conditional(lambda _: caching_models_enabled())


def create_workflow(
    description: str,
    initial_input_form: InputStepFunc | None = None,
    status: SubscriptionLifecycle = SubscriptionLifecycle.ACTIVE,
    additional_steps: StepList | None = None,
) -> Callable[[Callable[[], StepList]], Workflow]:
    """Transform an initial_input_form and a step list into a workflow with a target=Target.CREATE.

    Use this for create workflows only.

    Example::

        @create_workflow("create service port")
        def create_service_port() -> StepList:
            do_something
            >> do_something_else
    """
    create_initial_input_form_generator = wrap_create_initial_input_form(initial_input_form)

    def _create_workflow(f: Callable[[], StepList]) -> Workflow:
        steplist = (
            init
            >> f()
            >> (additional_steps or StepList())
            >> set_status(status)
            >> resync
            >> push_domain_models(cache_domain_models)
            >> refresh_subscription_search_index
            >> done
        )

        return make_workflow(f, description, create_initial_input_form_generator, Target.CREATE, steplist)

    return _create_workflow


def modify_workflow(
    description: str,
    initial_input_form: InputStepFunc | None = None,
    additional_steps: StepList | None = None,
) -> Callable[[Callable[[], StepList]], Workflow]:
    """Transform an initial_input_form and a step list into a workflow.

    Use this for modify workflows.

    Example::

        @modify_workflow("modify service port") -> StepList:
        def modify_service_port():
            do_something
            >> do_something_else
    """

    wrapped_modify_initial_input_form_generator = wrap_modify_initial_input_form(initial_input_form)

    def _modify_workflow(f: Callable[[], StepList]) -> Workflow:
        steplist = (
            init
            >> store_process_subscription(Target.MODIFY)
            >> unsync
            >> push_domain_models(cache_domain_models)
            >> f()
            >> (additional_steps or StepList())
            >> resync
            >> push_domain_models(cache_domain_models)
            >> refresh_subscription_search_index
            >> done
        )

        return make_workflow(f, description, wrapped_modify_initial_input_form_generator, Target.MODIFY, steplist)

    return _modify_workflow


def terminate_workflow(
    description: str,
    initial_input_form: InputStepFunc | None = None,
    additional_steps: StepList | None = None,
) -> Callable[[Callable[[], StepList]], Workflow]:
    """Transform an initial_input_form and a step list into a workflow.

    Use this for terminate workflows.

    Example::

        @terminate_workflow("terminate service port") -> StepList:
        def terminate_service_port():
            do_something
            >> do_something_else
    """

    wrapped_terminate_initial_input_form_generator = wrap_modify_initial_input_form(initial_input_form)

    def _terminate_workflow(f: Callable[[], StepList]) -> Workflow:
        steplist = (
            init
            >> store_process_subscription(Target.TERMINATE)
            >> unsync
            >> push_domain_models(cache_domain_models)
            >> f()
            >> (additional_steps or StepList())
            >> set_status(SubscriptionLifecycle.TERMINATED)
            >> resync
            >> push_domain_models(cache_domain_models)
            >> refresh_subscription_search_index
            >> done
        )

        return make_workflow(f, description, wrapped_terminate_initial_input_form_generator, Target.TERMINATE, steplist)

    return _terminate_workflow


def validate_workflow(description: str) -> Callable[[Callable[[], StepList]], Workflow]:
    """Transform an initial_input_form and a step list into a workflow.

    Use this for subscription validate workflows.

    Example::

        @validate_workflow("create service port")
        def create_service_port():
            do_something
            >> do_something_else
    """

    def _validate_workflow(f: Callable[[], StepList]) -> Workflow:
        push_subscriptions = conditional(lambda _: app_settings.CACHE_DOMAIN_MODELS)
        steplist = (
            init
            >> store_process_subscription(Target.SYSTEM)
            >> unsync_unchecked
            >> push_subscriptions(cache_domain_models)
            >> f()
            >> resync
            >> push_subscriptions(cache_domain_models)
            >> done
        )

        return make_workflow(f, description, validate_initial_input_form_generator, Target.SYSTEM, steplist)

    return _validate_workflow


@step("Equalize workflow step count")
def obsolete_step() -> None:
    """Equalize workflow step counts.

    When changing existing workflows that might have suspended instances in production it is important
    to keep the exact same amount of steps. This step can be used to fill the gap when removing a step.
    """
    pass



import json
from http import HTTPStatus

import pytest

from orchestrator.settings import AppSettings
from test.unit_tests.config import GRAPHQL_ENDPOINT, GRAPHQL_HEADERS


def get_customers_query(
    first: int = 1,
    after: int = 0,
    filter_by: list[str] | None = None,
    sort_by: list[dict[str, str]] | None = None,
) -> bytes:
    query = """
query CustomerQuery($first: Int!, $after: Int!, $filterBy: [GraphqlFilter!], $sortBy: [GraphqlSort!]) {
  customers(first: $first, after: $after, filterBy: $filterBy, sortBy: $sortBy) {
    page {
      customerId,
      fullname,
      shortcode,
    }
  }
}
    """
    return json.dumps(
        {
            "operationName": "CustomerQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
            },
        }
    ).encode("utf-8")


def test_customers(fastapi_app_graphql, test_client):
    response = test_client.post(GRAPHQL_ENDPOINT, content=get_customers_query(), headers=GRAPHQL_HEADERS)
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    customer_data = result["data"]["customers"]["page"][0]
    assert customer_data == {
        "customerId": "59289a57-70fb-4ff5-9c93-10fe67b12434",
        "fullname": "Default::Orchestrator-Core Customer",
        "shortcode": "default-cust",
    }


def test_change_customer_env_vars(fastapi_app_graphql, test_client):
    with pytest.MonkeyPatch.context() as mp:
        mp.setenv("DEFAULT_CUSTOMER_FULLNAME", "Custom Default Customer")
        mp.setenv("DEFAULT_CUSTOMER_SHORTCODE", "shortcode")
        mp.setenv("DEFAULT_CUSTOMER_IDENTIFIER", "123456")
        new_app_settings = AppSettings()
        mp.setattr("orchestrator.graphql.resolvers.customer.app_settings", new_app_settings)
        response = test_client.post(GRAPHQL_ENDPOINT, content=get_customers_query(), headers=GRAPHQL_HEADERS)

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    customer_data = result["data"]["customers"]["page"][0]
    assert customer_data == {
        "customerId": "123456",
        "fullname": "Custom Default Customer",
        "shortcode": "shortcode",
    }



from collections.abc import Sequence

from sqlalchemy import CompoundSelect, Select, select
from sqlalchemy.orm.strategy_options import _AbstractLoad

from orchestrator.db import db
from orchestrator.db.database import BaseModel


def rows_from_statement(
    stmt: Select | CompoundSelect,
    base_table: type[BaseModel],
    unique: bool = False,
    loaders: Sequence[_AbstractLoad] = (),
) -> Sequence:
    """Helper function to handle some tricky cases with sqlalchemy types."""
    # Tell SQLAlchemy that the rows must be objects of type `base_table` for CompoundSelect
    from_stmt = select(base_table).options(*loaders).from_statement(stmt)
    result = db.session.scalars(from_stmt)
    uresult = result.unique() if unique else result
    return uresult.all()



from typing import TYPE_CHECKING, Annotated

import strawberry
from strawberry.federation.schema_directives import Key
from strawberry.scalars import JSON
from strawberry.unset import UNSET

from oauth2_lib.strawberry import authenticated_field
from orchestrator.db import ProcessTable, ProductTable, db
from orchestrator.graphql.pagination import EMPTY_PAGE, Connection
from orchestrator.graphql.schemas.customer import CustomerType
from orchestrator.graphql.schemas.helpers import get_original_model
from orchestrator.graphql.schemas.product import ProductType
from orchestrator.graphql.types import GraphqlFilter, GraphqlSort, OrchestratorInfo
from orchestrator.schemas.process import ProcessSchema, ProcessStepSchema
from orchestrator.settings import app_settings

if TYPE_CHECKING:
    from orchestrator.graphql.schemas.subscription import SubscriptionInterface
else:
    SubscriptionInterface = Annotated["SubscriptionInterface", strawberry.lazy(".subscription")]

federation_key_directives = [Key(fields="processId", resolvable=UNSET)]


@strawberry.experimental.pydantic.type(model=ProcessStepSchema)
class ProcessStepType:
    step_id: strawberry.auto
    name: strawberry.auto
    status: strawberry.auto
    created_by: strawberry.auto
    executed: strawberry.auto
    commit_hash: strawberry.auto
    state: JSON | None
    state_delta: JSON | None


@strawberry.experimental.pydantic.type(model=ProcessSchema, directives=federation_key_directives)
class ProcessType:
    process_id: strawberry.auto
    product_id: strawberry.auto
    customer_id: strawberry.auto
    workflow_id: strawberry.auto
    workflow_name: strawberry.auto
    workflow_target: strawberry.auto
    assignee: strawberry.auto
    failed_reason: strawberry.auto
    last_step: strawberry.auto
    last_status: strawberry.auto
    created_by: strawberry.auto
    started_at: strawberry.auto
    last_modified_at: strawberry.auto
    is_task: strawberry.auto
    steps: strawberry.auto
    form: JSON | None
    current_state: JSON | None

    @strawberry.field(description="Get traceback")  # type: ignore
    def traceback(self) -> str | None:
        model = get_original_model(self, ProcessTable)
        return model.traceback

    @authenticated_field(description="Returns the associated product")  # type: ignore
    def product(self) -> ProductType | None:
        if self.product_id and (product := db.session.get(ProductTable, self.product_id)):
            return ProductType.from_pydantic(product)
        return None

    @strawberry.field(description="Returns customer of a subscription")  # type: ignore
    def customer(self) -> CustomerType:
        return CustomerType(
            customer_id=app_settings.DEFAULT_CUSTOMER_IDENTIFIER,
            fullname=app_settings.DEFAULT_CUSTOMER_FULLNAME,
            shortcode=app_settings.DEFAULT_CUSTOMER_SHORTCODE,
        )

    @authenticated_field(description="Returns list of subscriptions of the process")  # type: ignore
    async def subscriptions(
        self,
        info: OrchestratorInfo,
        filter_by: list[GraphqlFilter] | None = None,
        sort_by: list[GraphqlSort] | None = None,
        first: int = 10,
        after: int = 0,
    ) -> Connection[SubscriptionInterface]:
        from orchestrator.graphql.resolvers.subscription import resolve_subscriptions

        subscription_ids = [str(s.subscription_id) for s in self._original_model.subscriptions]  # type: ignore
        if not subscription_ids:
            return EMPTY_PAGE
        filter_by_with_related_subscriptions = (filter_by or []) + [
            GraphqlFilter(field="subscriptionId", value="|".join(subscription_ids))
        ]
        return await resolve_subscriptions(info, filter_by_with_related_subscriptions, sort_by, first, after)



# Copyright 2019-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum, EnumMeta
from typing import Any, get_args

import strawberry
import structlog
from more_itertools import one
from strawberry.experimental.pydantic.conversion_types import StrawberryTypeFromPydantic
from strawberry.federation.schema_directives import Key
from strawberry.unset import UNSET

from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import DomainModel, get_depends_on_product_block_type_list
from orchestrator.graphql.schemas.product_block import BaseProductBlockType
from orchestrator.graphql.types import StrawberryModelType
from orchestrator.types import filter_nonetype, is_of_type, is_optional_type
from orchestrator.utils.helpers import to_camel

logger = structlog.get_logger(__name__)

EnumDict = dict[str, EnumMeta]

# Mapping to specifically prevent aliases for fields from being used when creating strawberry types.
# use like `USE_PYDANTIC_ALIAS_MODEL_MAPPING.updates({ "ProductNameSubscription": False })`, default is True.
# more info can be read here: https://workfloworchestrator.org/orchestrator-core/architecture/application/graphql/#usage-of-use-pydantic-alias-model-mapping
USE_PYDANTIC_ALIAS_MODEL_MAPPING: dict[str, bool] = {}


def create_strawberry_enum(enum: Any) -> EnumMeta:
    return strawberry.enum(enum)


def is_not_strawberry_enum(key: str, strawberry_enums: EnumDict) -> bool:
    return key not in strawberry_enums


def _get_enum(field: Any) -> Enum | None:
    if is_optional_type(field, Enum):
        return one(filter_nonetype(get_args(field)))
    if is_of_type(field, Enum):
        return field
    return None


def create_strawberry_enums(model: type[DomainModel], strawberry_enums: EnumDict) -> EnumDict:
    enums = {
        key: create_strawberry_enum(field)
        for key, raw_field in model._non_product_block_fields_.items()
        if (field := _get_enum(raw_field)) and is_not_strawberry_enum(key, strawberry_enums)
    }
    return strawberry_enums | enums


def graphql_name(name: str) -> str:
    return to_camel(name.replace(" ", "_"))


def graphql_subscription_name(name: str) -> str:
    subscription_graphql_name = graphql_name(name).replace("Initial", "").replace("Inactive", "")
    return f"{subscription_graphql_name}Subscription"


def create_block_strawberry_type(
    strawberry_name: str,
    model: type[DomainModel],
) -> type[StrawberryTypeFromPydantic[DomainModel]]:
    from strawberry.federation.schema_directives import Key
    from strawberry.unset import UNSET

    federation_key_directives = [Key(fields="subscriptionInstanceId", resolvable=UNSET)]

    if keys := [key for key in model.__annotations__.keys() if "_id" in key]:
        federation_key_directives.extend([Key(fields=to_camel(key), resolvable=True) for key in keys])

    base_type = type(strawberry_name, (BaseProductBlockType,), {})
    pydantic_wrapper = strawberry.experimental.pydantic.type(
        model,
        all_fields=True,
        directives=federation_key_directives,
        description=f"{strawberry_name} Type",
        use_pydantic_alias=USE_PYDANTIC_ALIAS_MODEL_MAPPING.get(strawberry_name, True),
    )
    return type(strawberry_name, (pydantic_wrapper(base_type),), {})


def create_subscription_strawberry_type(strawberry_name: str, model: type[DomainModel], interface: type) -> type:
    base_type = type(strawberry_name, (interface,), {})
    directives = [Key(fields="subscriptionId", resolvable=UNSET)]

    pydantic_wrapper = strawberry.experimental.pydantic.type(
        model,
        all_fields=True,
        directives=directives,
        description=f"{strawberry_name} Type",
        use_pydantic_alias=USE_PYDANTIC_ALIAS_MODEL_MAPPING.get(strawberry_name, True),
    )
    return type(strawberry_name, (pydantic_wrapper(base_type),), {})


def add_class_to_strawberry(
    model_name: str,
    model: type[DomainModel],
    strawberry_models: StrawberryModelType,
    strawberry_enums: EnumDict,
    interface: type | None = None,
) -> None:
    if model_name in strawberry_models:
        logger.debug("Skip already registered strawberry model", model=repr(model), strawberry_name=model_name)
        return
    logger.debug("Registering strawberry model", model=repr(model), strawberry_name=model_name)

    strawberry_enums = create_strawberry_enums(model, strawberry_enums)

    product_blocks_types_in_model = get_depends_on_product_block_type_list(model._get_depends_on_product_block_types())
    for field in product_blocks_types_in_model:
        graphql_field_name = graphql_name(field.__name__)
        if graphql_field_name not in strawberry_models and graphql_field_name != model_name:
            add_class_to_strawberry(graphql_field_name, field, strawberry_models, strawberry_enums)

    if interface:
        strawberry_models[model_name] = create_subscription_strawberry_type(model_name, model, interface)
    else:
        strawberry_models[model_name] = create_block_strawberry_type(model_name, model)
    logger.debug("Registered strawberry model", model=repr(model), strawberry_name=model_name)


def register_domain_models(
    interface: type | None, existing_models: StrawberryModelType | None = None
) -> StrawberryModelType:
    strawberry_models = existing_models if existing_models else {}
    strawberry_enums: EnumDict = {}
    products = {
        product_type.__base_type__.__name__: product_type.__base_type__
        for product_type in SUBSCRIPTION_MODEL_REGISTRY.values()
        if product_type.__base_type__
    }
    for key, product_type in products.items():
        add_class_to_strawberry(
            interface=interface,
            model_name=graphql_subscription_name(key),
            model=product_type,
            strawberry_models=strawberry_models,
            strawberry_enums=strawberry_enums,
        )
    return strawberry_models



from typing import Iterable

import structlog
from sqlalchemy.orm import Load

from orchestrator.db.database import BaseModel as DbBaseModel
from orchestrator.db.loaders import AttrLoader, join_attr_loaders, lookup_attr_loaders
from orchestrator.graphql.types import OrchestratorInfo
from orchestrator.graphql.utils.get_selected_paths import get_selected_paths

logger = structlog.get_logger(__name__)


def _split_path(query_path: str) -> Iterable[str]:
    yield from (field for field in query_path.split(".") if field != "page")


def get_query_loaders(info: OrchestratorInfo, root_model: type[DbBaseModel]) -> list[Load]:
    """Get sqlalchemy query loaders for the given GraphQL query.

    Based on the GraphQL query's selected fields, returns the required DB loaders to use
    in SQLALchemy's `.options()` for efficiently quering (nested) relationships.
    """
    # Strip page and sort by length to find the longest match first
    query_paths = [path.removeprefix("page.") for path in get_selected_paths(info)]
    query_paths.sort(key=lambda x: x.count("."), reverse=True)

    def get_loader_for_path(query_path: str) -> tuple[str, Load | None]:
        next_model = root_model

        matched_fields: list[str] = []
        path_loaders: list[AttrLoader] = []

        for field in _split_path(query_path):
            if not (attr_loaders := lookup_attr_loaders(next_model, field)):
                break

            matched_fields.append(field)
            path_loaders.extend(attr_loaders)
            next_model = attr_loaders[-1].next_model

        return ".".join(matched_fields), join_attr_loaders(path_loaders)

    query_loaders: dict[str, Load] = {}

    for path in query_paths:
        matched_path, loader = get_loader_for_path(path)
        if not matched_path or not loader or matched_path in query_loaders:
            continue
        if any(known_path.startswith(f"{matched_path}.") for known_path in query_loaders):
            continue
        query_loaders[matched_path] = loader

    loaders = list(query_loaders.values())
    logger.debug(
        "Generated query loaders",
        root_model=root_model,
        query_paths=query_paths,
        query_loaders=[str(i.path) for i in loaders],
    )
    return loaders



import structlog
from orchestrator.types import State
from orchestrator.workflow import StepList, begin, step
from orchestrator.workflows.utils import validate_workflow

from products.product_types.example1 import Example1

logger = structlog.get_logger(__name__)


@step("Load initial state")
def load_initial_state_example1(subscription: Example1) -> State:
    return {
        "subscription": subscription,
    }


@step("Validate that the example1 subscription is correctly administered in some external system")
def check_validate_example_in_some_oss(subscription: Example1) -> State:
    # TODO: add validation for "Validate that the example1 subscription is correctly administered in some external system"
    if True:
        raise ValueError("Validate that the example1 subscription is correctly administered in some external system")

    return {}


@validate_workflow("Validate example1")
def validate_example1() -> StepList:
    return begin >> load_initial_state_example1 >> check_validate_example_in_some_oss



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from orchestrator.domain.base import SubscriptionModel
from orchestrator.utils.docs import make_product_type_index_doc

SUBSCRIPTION_MODEL_REGISTRY: dict[str, type[SubscriptionModel]] = {}

__doc__ = make_product_type_index_doc(SUBSCRIPTION_MODEL_REGISTRY)

__all__ = ["SubscriptionModel", "SUBSCRIPTION_MODEL_REGISTRY"]



from datetime import datetime
from unittest import mock
from uuid import uuid4

import pytest
import pytz
from dirty_equals import IsUUID
from pydantic import Field, ValidationError, computed_field, conlist
from pydantic.main import BaseModel
from sqlalchemy import func, select
from sqlalchemy.exc import NoResultFound

from orchestrator.db import (
    ProductTable,
    SubscriptionInstanceRelationTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    db,
)
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import (
    DomainModel,
    ProductBlockModel,
    SubscriptionModel,
)
from orchestrator.domain.lifecycle import ProductLifecycle
from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.fixtures.products.product_blocks.product_block_list_nested import (
    ProductBlockListNestedForTestInactive,
)
from test.unit_tests.fixtures.products.product_blocks.product_block_one import DummyEnum
from test.unit_tests.fixtures.products.product_blocks.product_block_one_nested import (
    ProductBlockOneNestedForTestInactive,
)


def get_one_relation(depends_on_id):
    return db.session.scalars(
        select(SubscriptionInstanceRelationTable).where(
            SubscriptionInstanceRelationTable.depends_on_id == depends_on_id
        )
    ).one()


def test_product_block_metadata(test_product_block_one, test_product_one, test_product_block_one_db):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    subscription_id = uuid4()
    ProductBlockOneForTestInactive.new(
        subscription_id=subscription_id
    )  # Need at least one instance since we lazy load this stuff

    product_block, _ = test_product_block_one_db

    assert ProductBlockOneForTestInactive.name == "ProductBlockOneForTest"
    assert ProductBlockOneForTestInactive.description == "Test Block"
    assert ProductBlockOneForTestInactive.product_block_id == product_block.product_block_id
    assert ProductBlockOneForTestInactive.tag == "TEST"


def test_product_block_one_nested(test_product_model_nested, test_product_type_one_nested):
    """Test the behavior of nesting (self-referencing) product blocks.

    Notes:
        - nesting only works when each block is attached to a different subscription

    """
    ProductTypeOneNestedForTestInactive, _, ProductTypeOneNestedForTest = test_product_type_one_nested
    customer_id = str(uuid4())

    def create_subscription(*, int_value, sub_block=None):
        subscription = ProductTypeOneNestedForTestInactive.from_product_id(
            product_id=test_product_model_nested.product_id, customer_id=customer_id, insync=True
        )
        subscription.block = ProductBlockOneNestedForTestInactive.new(
            subscription_id=subscription.subscription_id, int_field=int_value, sub_block=sub_block
        )
        subscription = SubscriptionModel.from_other_lifecycle(subscription, SubscriptionLifecycle.ACTIVE)
        subscription.save()
        db.session.commit()
        return subscription

    def assert_int_values(subscription, *, level1, level2, level3):
        block = subscription.block
        actual_values = (block.int_field, block.sub_block.int_field, block.sub_block.sub_block.int_field)
        assert (level1, level2, level3) == actual_values
        assert block.sub_block.sub_block.sub_block is None

    # Create productblock 30 that will be nested in block 20
    subscription_30 = create_subscription(int_value=30)

    # Create productblock 20 that refers to block 30, and will be nested in block 10
    subscription_20 = create_subscription(int_value=20, sub_block=subscription_30.block)

    # Create productblock 10 that refers to block 20
    subscription_10 = create_subscription(int_value=10, sub_block=subscription_20.block)

    # Load block 10 and verify the nested blocks
    subscription_10 = ProductTypeOneNestedForTest.from_subscription(subscription_10.subscription_id)
    assert_int_values(subscription_10, level1=10, level2=20, level3=30)

    # Load block 20 and verify the nested block
    subscription_20 = ProductTypeOneNestedForTest.from_subscription(subscription_20.subscription_id)
    assert subscription_20.block.int_field == 20
    assert subscription_20.block.sub_block.int_field == 30
    assert subscription_20.block.sub_block.sub_block is None

    # Create productblock 11 that also refers to block 20
    subscription_11 = create_subscription(int_value=11, sub_block=subscription_20.block)

    # Load block 11 and verify the nested blocks
    subscription_11 = ProductTypeOneNestedForTest.from_subscription(subscription_11.subscription_id)
    assert_int_values(subscription_11, level1=11, level2=20, level3=30)

    # (again) Load block 10 and verify the nested blocks are same as before
    subscription_10 = ProductTypeOneNestedForTest.from_subscription(subscription_10.subscription_id)
    assert_int_values(subscription_10, level1=10, level2=20, level3=30)

    # Below part might not be interesting to test, or better off in a separate testcase.
    # I was just curious what happens when we delete things.

    # Remove block 20 from block 10
    subscription_10.block.sub_block = None
    subscription_10.save()
    db.session.commit()

    # Load block 10 and verify the nested block is removed
    subscription_10 = ProductTypeOneNestedForTest.from_subscription(subscription_10.subscription_id)
    assert subscription_10.block.int_field == 10
    assert subscription_10.block.sub_block is None

    # Load block 11 and verify the nested blocks still exist
    subscription_11 = ProductTypeOneNestedForTest.from_subscription(subscription_11.subscription_id)
    assert_int_values(subscription_11, level1=11, level2=20, level3=30)


def test_product_block_list_nested(test_product_model_list_nested, test_product_type_list_nested):
    """Test the behavior of nesting (self-referencing) a list of product blocks.

    Notes:
        - nesting only works when each block is attached to a different subscription

    """
    ProductTypeListNestedForTestInactive, _, ProductTypeListNestedForTest = test_product_type_list_nested

    customer_id = str(uuid4())

    def create_subscription(*, int_value, sub_blocks=()):
        subscription = ProductTypeListNestedForTestInactive.from_product_id(
            product_id=test_product_model_list_nested.product_id, customer_id=customer_id, insync=True
        )
        subscription.block = ProductBlockListNestedForTestInactive.new(
            subscription_id=subscription.subscription_id, int_field=int_value, sub_block_list=list(sub_blocks)
        )
        subscription = SubscriptionModel.from_other_lifecycle(subscription, SubscriptionLifecycle.ACTIVE)
        subscription.save()
        db.session.commit()
        return subscription

    def assert_int_values(subscription, *, level1, level2, level3):
        block = subscription.block
        assert level1 == block.int_field
        assert level2 == sorted(l2.int_field for l2 in block.sub_block_list)
        assert level3 == sorted(l3.int_field for l2 in block.sub_block_list for l3 in l2.sub_block_list)

    # Create productblocks 30 and 31 that will both be nested in block 20 and 21
    subscription_30 = create_subscription(int_value=30)
    subscription_31 = create_subscription(int_value=31)

    # Create productblocks 20 and 21 that both
    # - refer to blocks 30 and 31
    # - will be nested in blocks 10 and 11
    subscription_20 = create_subscription(int_value=20, sub_blocks=(subscription_30.block, subscription_31.block))
    subscription_21 = create_subscription(int_value=21, sub_blocks=(subscription_30.block, subscription_31.block))

    # Create productblocks 10 and 11 that both refer to blocks 20 and 21
    subscription_10 = create_subscription(int_value=10, sub_blocks=(subscription_20.block, subscription_21.block))
    subscription_11 = create_subscription(int_value=11, sub_blocks=(subscription_20.block, subscription_21.block))

    # Load blocks 10 and 11 and verify the nested blocks
    subscription_10 = ProductTypeListNestedForTest.from_subscription(subscription_10.subscription_id)
    subscription_11 = ProductTypeListNestedForTest.from_subscription(subscription_11.subscription_id)
    assert_int_values(subscription_10, level1=10, level2=[20, 21], level3=[30, 30, 31, 31])
    assert_int_values(subscription_11, level1=11, level2=[20, 21], level3=[30, 30, 31, 31])

    # Assert that blocks at deepest level don't refer to any other blocks
    assert subscription_10.block.sub_block_list[0].sub_block_list[0].sub_block_list == []
    assert subscription_11.block.sub_block_list[1].sub_block_list[1].sub_block_list == []

    # Load block 20 and verify nested blocks
    subscription_20 = ProductTypeListNestedForTest.from_subscription(subscription_20.subscription_id)
    assert_int_values(subscription_20, level1=20, level2=[30, 31], level3=[])

    # Create productblock 32 and nest it in block 20 (but not 21!)
    subscription_32 = create_subscription(int_value=32)
    subscription_20.block.sub_block_list.append(subscription_32.block)
    subscription_20.save()
    db.session.commit()

    # (again) Load blocks 10 and 11 and verify block 32 is present once
    subscription_10 = ProductTypeListNestedForTest.from_subscription(subscription_10.subscription_id)
    subscription_11 = ProductTypeListNestedForTest.from_subscription(subscription_11.subscription_id)
    assert_int_values(subscription_10, level1=10, level2=[20, 21], level3=[30, 30, 31, 31, 32])
    assert_int_values(subscription_11, level1=11, level2=[20, 21], level3=[30, 30, 31, 31, 32])


def test_lifecycle(test_product_model, test_product_type_one, test_product_block_one):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one
    ProductTypeOneForTestInactive, _, _ = test_product_type_one
    subscription_id = uuid4()

    # Test create with wrong lifecycle, we can create
    with pytest.raises(ValueError, match=r"is not valid for status active"):
        ProductTypeOneForTestInactive(
            product=test_product_model,
            customer_id=str(uuid4()),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            block=ProductBlockOneForTestInactive.new(subscription_id),
            status=SubscriptionLifecycle.ACTIVE,
            test_fixed_input=False,
        )

    # Works with right lifecycle
    product_type = ProductTypeOneForTestInactive(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTestInactive.new(subscription_id),
        status=SubscriptionLifecycle.INITIAL,
        test_fixed_input=False,
    )

    assert product_type.status == SubscriptionLifecycle.INITIAL


def test_lifecycle_specific(
    test_product_model, test_product_type_one, test_product_block_one, test_product_sub_block_one
):
    _, SubBlockOneForTestProvisioning, SubBlockOneForTest = test_product_sub_block_one
    _, ProductBlockOneForTestProvisioning, ProductBlockOneForTest = test_product_block_one
    _, ProductTypeOneForTestProvisioning, ProductTypeOneForTest = test_product_type_one
    subscription_id = uuid4()

    # Works with less contrained lifecycle
    product_type = ProductTypeOneForTest(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            enum_field=DummyEnum.FOO,
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.INITIAL,
        test_fixed_input=False,
    )

    assert product_type.status == SubscriptionLifecycle.INITIAL

    # Works with right lifecycle
    product_type = ProductTypeOneForTest(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            enum_field=DummyEnum.FOO,
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.ACTIVE,
        test_fixed_input=False,
    )

    assert product_type.status == SubscriptionLifecycle.ACTIVE

    # Does not work with more constrained lifecycle
    with pytest.raises(ValueError, match=r"is not valid for status active"):
        ProductTypeOneForTestProvisioning(
            product=test_product_model,
            customer_id=str(uuid4()),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            block=ProductBlockOneForTestProvisioning.new(
                subscription_id=subscription_id,
                int_field=3,
                list_field=[1],
                enum_field=DummyEnum.FOO,
                sub_block=SubBlockOneForTestProvisioning.new(subscription_id=subscription_id, int_field=3),
                sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            ),
            status=SubscriptionLifecycle.ACTIVE,
            test_fixed_input=False,
        )

    # Works with right lifecycle
    product_type = ProductTypeOneForTestProvisioning(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTestProvisioning.new(
            subscription_id=subscription_id,
            int_field=3,
            list_field=[1],
            enum_field=DummyEnum.FOO,
            sub_block=SubBlockOneForTestProvisioning.new(subscription_id=subscription_id, int_field=3),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.PROVISIONING,
        test_fixed_input=False,
    )
    assert product_type.status == SubscriptionLifecycle.PROVISIONING


def test_product_blocks_per_lifecycle(
    test_product_model, test_product_type_one, test_product_block_one, test_product_sub_block_one
):
    _, _, SubBlockOneForTest = test_product_sub_block_one
    ProductBlockOneForTestInactive, _, ProductBlockOneForTest = test_product_block_one
    ProductTypeOneForTestInactive, ProductTypeOneForTestProvisioning, ProductTypeOneForTest = test_product_type_one
    subscription_id = uuid4()

    ProductTypeOneForTestInactive(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            enum_field=DummyEnum.FOO,
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.INITIAL,
        test_fixed_input=False,
    )

    ProductTypeOneForTestInactive(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTestInactive.new(subscription_id=subscription_id, int_field=3),
        status=SubscriptionLifecycle.INITIAL,
        test_fixed_input=False,
    )

    ProductTypeOneForTestProvisioning(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            enum_field=DummyEnum.FOO,
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.PROVISIONING,
        test_fixed_input=False,
    )

    ProductTypeOneForTest(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            enum_field=DummyEnum.FOO,
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.ACTIVE,
        test_fixed_input=False,
    )

    ProductTypeOneForTest(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            enum_field=DummyEnum.FOO,
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.INITIAL,
        test_fixed_input=False,
    )

    with pytest.raises(
        ValidationError,
        match=(
            r"2 validation errors for SubBlockOneForTest\nint_field\n\s+Field required.+\n.+\nstr_field\n\s+Field required"
        ),  # Note: the . does not match newlines
    ):
        ProductTypeOneForTest(
            product=test_product_model,
            customer_id=str(uuid4()),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            block=ProductBlockOneForTest.new(subscription_id=subscription_id),
            status=SubscriptionLifecycle.ACTIVE,
            test_fixed_input=False,
        )

    with pytest.raises(
        ValidationError,
        match=r"1 validation error for ProductTypeOneForTest\nblock\n.+dictionary or instance of ProductBlockOneForTest",
    ):
        ProductTypeOneForTest(
            product=test_product_model,
            customer_id=str(uuid4()),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            block=ProductBlockOneForTestInactive.new(subscription_id=subscription_id),
            status=SubscriptionLifecycle.ACTIVE,
            test_fixed_input=False,
        )

    with pytest.raises(
        ValidationError, match=r"1 validation error for ProductTypeOneForTest\nblock\n\s+Field required .+"
    ):
        ProductTypeOneForTest(
            product=test_product_model,
            customer_id=str(uuid4()),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            status=SubscriptionLifecycle.INITIAL,
            test_fixed_input=False,
        )

    with pytest.raises(
        ValidationError, match=r"1 validation error for ProductTypeOneForTest\nblock\n\s+Field required .+"
    ):
        ProductTypeOneForTest(
            product=test_product_model,
            customer_id=str(uuid4()),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            status=SubscriptionLifecycle.ACTIVE,
            test_fixed_input=False,
        )


def test_change_lifecycle(test_product_one, test_product_type_one, test_product_block_one, test_product_sub_block_one):
    SubBlockOneForTestInactive, _, _ = test_product_sub_block_one
    ProductBlockOneForTestInactive, _, _ = test_product_block_one
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    product_type = ProductTypeOneForTestInactive.from_product_id(
        test_product_one,
        str(uuid4()),
    )
    product_type.block = ProductBlockOneForTestInactive.new(
        subscription_id=product_type.subscription_id,
        sub_block_2=SubBlockOneForTestInactive.new(subscription_id=product_type.subscription_id),
    )

    # Does not work if constraints are not met
    with pytest.raises(
        ValidationError,
        match=r"int_field\n\s+Input should be a valid integer.+\n.+\nstr_field\n\s+Input should be a valid string",
    ):
        product_type = SubscriptionModel.from_other_lifecycle(product_type, SubscriptionLifecycle.ACTIVE)
    with pytest.raises(ValidationError, match=r"int_field\n\s+Input should be a valid integer"):
        product_type = SubscriptionModel.from_other_lifecycle(product_type, SubscriptionLifecycle.PROVISIONING)

    # Set first value
    product_type.block.int_field = 3
    product_type.block.sub_block.int_field = 3
    product_type.block.sub_block_2.int_field = 3
    product_type.block.sub_block_list = [SubBlockOneForTestInactive.new(subscription_id=product_type.subscription_id)]
    product_type.block.sub_block_list[0].int_field = 4
    product_type.block.list_field = [1]
    product_type.block.enum_field = DummyEnum.FOO

    # Does not work if constraints are not met
    with pytest.raises(ValidationError, match=r"str_field\n\s+Input should be a valid string"):
        product_type = SubscriptionModel.from_other_lifecycle(product_type, SubscriptionLifecycle.ACTIVE)

    # works with correct data
    product_type = SubscriptionModel.from_other_lifecycle(product_type, SubscriptionLifecycle.PROVISIONING)
    assert product_type.status == SubscriptionLifecycle.PROVISIONING
    assert product_type.block.str_field is None

    product_type.block.str_field = "A"
    product_type.block.sub_block.str_field = "B"
    product_type.block.sub_block_2.str_field = "C"
    product_type.block.sub_block_list[0].str_field = "D"

    # works with correct data
    product_type_new = SubscriptionModel.from_other_lifecycle(product_type, SubscriptionLifecycle.ACTIVE)
    assert product_type_new.status == SubscriptionLifecycle.ACTIVE
    expected_dict = product_type.model_dump()
    expected_dict["status"] = SubscriptionLifecycle.ACTIVE
    expected_dict["start_date"] = mock.ANY
    assert product_type_new.model_dump() == expected_dict
    assert isinstance(product_type_new.start_date, datetime)


def test_save_load(test_product_model, test_product_type_one, test_product_block_one, test_product_sub_block_one):
    SubBlockOneForTestInactive, SubBlockOneForTestProvisioning, SubBlockOneForTest = test_product_sub_block_one
    ProductBlockOneForTestInactive, ProductBlockOneForTestProvisioning, ProductBlockOneForTest = test_product_block_one
    ProductTypeOneForTestInactive, ProductTypeOneForTestProvisioning, ProductTypeOneForTest = test_product_type_one

    customer_id = str(uuid4())

    model = ProductTypeOneForTestInactive.from_product_id(
        product_id=test_product_model.product_id,
        customer_id=customer_id,
        insync=True,
        description="Desc",
        start_date=datetime(2021, 1, 1, 1, 1, 1, tzinfo=pytz.utc),
        end_date=datetime(2021, 1, 1, 1, 1, 2, tzinfo=pytz.utc),
        note="Note",
        status=SubscriptionLifecycle.INITIAL,
    )

    # Set one value and make sure its saved
    model.block.str_field = "A"

    model.save()
    db.session.commit()

    assert (
        db.session.scalar(
            select(func.count())
            .select_from(SubscriptionInstanceValueTable)
            .join(SubscriptionInstanceValueTable.subscription_instance)
            .filter(SubscriptionInstanceTable.subscription_id == model.subscription_id)
        )
        == 1
    )

    assert model.model_dump() == {
        "block": {
            "int_field": None,
            "enum_field": None,
            "label": None,
            "list_field": [],
            "title": "TEST ProductBlockOneForTestInactive int_field=None",
            "name": "ProductBlockOneForTest",
            "str_field": "A",
            "sub_block": {
                "int_field": None,
                "label": None,
                "name": "SubBlockOneForTest",
                "str_field": None,
                "subscription_instance_id": IsUUID(),
                "owner_subscription_id": IsUUID(),
            },
            "sub_block_2": None,
            "sub_block_list": [],
            "owner_subscription_id": IsUUID(),
            "subscription_instance_id": IsUUID(),
        },
        "customer_id": customer_id,
        "description": "Desc",
        "end_date": datetime(2021, 1, 1, 1, 1, 2, tzinfo=pytz.utc),
        "insync": True,
        "note": "Note",
        "product": {
            "description": "Test ProductTable",
            "name": "TestProductOne",
            "product_id": test_product_model.product_id,
            "product_type": "Test",
            "status": ProductLifecycle.ACTIVE,
            "tag": "TEST",
            "created_at": test_product_model.created_at,
            "end_date": None,
        },
        "start_date": datetime(2021, 1, 1, 1, 1, 1, tzinfo=pytz.utc),
        "status": SubscriptionLifecycle.INITIAL,
        "subscription_id": IsUUID(),
        "test_fixed_input": False,
    }

    # Set first value
    model.block.int_field = 3
    model.block.enum_field = DummyEnum.FOO
    model.block.sub_block.int_field = 3
    model.block.sub_block_2 = SubBlockOneForTestInactive.new(subscription_id=model.subscription_id)
    model.block.sub_block_2.int_field = 3
    model.block.list_field = [1]
    model.block.sub_block.str_field = "B"
    model.block.sub_block_2.str_field = "C"

    # works with correct data
    model = SubscriptionModel.from_other_lifecycle(model, SubscriptionLifecycle.ACTIVE)

    model.save()
    db.session.commit()

    new_model = ProductTypeOneForTest.from_subscription(model.subscription_id)
    assert model.model_dump() == new_model.model_dump()

    # Second save also works as expected
    new_model.save()
    db.session.commit()

    latest_model = ProductTypeOneForTest.from_subscription(model.subscription_id)
    assert new_model.model_dump() == latest_model.model_dump()

    # Loading blocks also works
    block = ProductBlockOneForTest.from_db(model.block.subscription_instance_id)
    assert block.model_dump() == model.block.model_dump()


def test_update_constrained_lists(test_product_one, test_product_block_one):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    class TestConListProductType(SubscriptionModel, is_base=True):
        saps: conlist(ProductBlockOneForTestInactive, min_length=1, max_length=4)

    # Creates
    ip = TestConListProductType.from_product_id(product_id=test_product_one, customer_id=str(uuid4()))
    ip.save()

    sap = ProductBlockOneForTestInactive.new(subscription_id=ip.subscription_id, int_field=3, str_field="")

    # Set new saps, removes old one
    ip.saps = [sap]

    ip.save()

    ip2 = TestConListProductType.from_subscription(ip.subscription_id)

    # Old default saps should not be saved
    assert ip.model_dump() == ip2.model_dump()

    # Test constraint
    with pytest.raises(ValidationError):
        ip.saps = []


def test_update_lists(test_product_one, test_product_block_one):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    class TestListProductType(SubscriptionModel, is_base=True):
        saps: list[ProductBlockOneForTestInactive]

    # Creates
    ip = TestListProductType.from_product_id(product_id=test_product_one, customer_id=str(uuid4()))
    ip.save()

    sap = ProductBlockOneForTestInactive.new(subscription_id=ip.subscription_id, int_field=3, str_field="")

    # Set new saps
    ip.saps = [sap]

    ip.save()

    ip2 = TestListProductType.from_subscription(ip.subscription_id)

    # Old default saps should not be saved
    assert ip.model_dump() == ip2.model_dump()


def test_update_optional(test_product_one, test_product_block_one):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    class TestListProductType(SubscriptionModel, is_base=True):
        sap: ProductBlockOneForTestInactive | None = None

    # Creates
    ip = TestListProductType.from_product_id(product_id=test_product_one, customer_id=str(uuid4()))
    ip.save()

    sap = ProductBlockOneForTestInactive.new(subscription_id=ip.subscription_id, int_field=3, str_field="")

    # Set new sap
    ip.sap = sap

    ip.save()

    ip2 = TestListProductType.from_subscription(ip.subscription_id)

    # Old default saps should not be saved
    assert ip.model_dump() == ip2.model_dump()


def test_generic_from_subscription(test_product_one, test_product_type_one):
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    ip = ProductTypeOneForTestInactive.from_product_id(product_id=test_product_one, customer_id=str(uuid4()))
    ip.save()

    model = SubscriptionModel.from_subscription(ip.subscription_id)

    assert isinstance(model, ProductTypeOneForTestInactive)


def test_label_is_saved(test_product_one, test_product_type_one):
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    test_model = ProductTypeOneForTestInactive.from_product_id(test_product_one, str(uuid4()))
    test_model.block.label = "My label"
    test_model.save()
    db.session.commit()
    instance_in_db = db.session.get(SubscriptionInstanceTable, test_model.block.subscription_instance_id)
    assert instance_in_db.label == "My label"


def test_domain_model_attrs_saving_loading(test_product_one, test_product_type_one, test_product_sub_block_one):
    SubBlockOneForTestInactive, _, _ = test_product_sub_block_one
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    test_model = ProductTypeOneForTestInactive.from_product_id(product_id=test_product_one, customer_id=str(uuid4()))
    test_model.block.sub_block_2 = SubBlockOneForTestInactive.new(subscription_id=test_model.subscription_id)
    test_model.block.sub_block_list = [SubBlockOneForTestInactive.new(subscription_id=test_model.subscription_id)]
    test_model.save()
    db.session.commit()

    def get_one_relation(depends_on_id):
        return db.session.scalars(
            select(SubscriptionInstanceRelationTable).where(
                SubscriptionInstanceRelationTable.depends_on_id == depends_on_id
            )
        ).one()

    relation = get_one_relation(test_model.block.sub_block.subscription_instance_id)
    assert relation.domain_model_attr == "sub_block"

    relation = get_one_relation(test_model.block.sub_block_2.subscription_instance_id)
    assert relation.domain_model_attr == "sub_block_2"

    relation = get_one_relation(test_model.block.sub_block_list[0].subscription_instance_id)
    assert relation.domain_model_attr == "sub_block_list"

    test_model_2 = ProductTypeOneForTestInactive.from_subscription(test_model.subscription_id)
    assert test_model == test_model_2


def test_removal_of_domain_attrs(test_product_one, test_product_type_one, test_product_sub_block_one):
    SubBlockOneForTestInactive, _, _ = test_product_sub_block_one
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    test_model = ProductTypeOneForTestInactive.from_product_id(product_id=test_product_one, customer_id=str(uuid4()))
    test_model.block.sub_block_2 = SubBlockOneForTestInactive.new(subscription_id=test_model.subscription_id)

    test_model.save()
    db.session.commit()
    relation = get_one_relation(test_model.block.sub_block.subscription_instance_id)
    relation.domain_model_attr = None
    db.session.commit()
    with pytest.raises(ValueError, match=r"Expected exactly one item in iterable, but got"):
        ProductTypeOneForTestInactive.from_subscription(test_model.subscription_id)


def test_simple_model_with_no_attrs(generic_subscription_1, generic_product_type_1):
    GenericProductOneInactive, GenericProductOne = generic_product_type_1
    model = GenericProductOne.from_subscription(subscription_id=generic_subscription_1)
    with pytest.raises(NoResultFound):
        get_one_relation(model.pb_1.subscription_instance_id)


def test_abstract_super_block(test_product_one, test_product_type_one, test_product_block_one_db):
    ProductTypeOneForTestInactive, ProductTypeOneForTestProvisioning, ProductTypeOneForTest = test_product_type_one

    class AbstractProductBlockOneForTestInactive(ProductBlockModel):
        str_field: str | None = None
        list_field: list[int] = Field(default_factory=list)

    class AbstractProductBlockOneForTestProvisioning(
        AbstractProductBlockOneForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
    ):
        str_field: str | None = None
        list_field: list[int]

    class AbstractProductBlockOneForTest(
        AbstractProductBlockOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        str_field: str
        list_field: list[int]

    class ProductBlockOneForTestInactive(
        AbstractProductBlockOneForTestInactive, product_block_name="ProductBlockOneForTest"
    ):
        str_field: str | None = None
        list_field: list[int] = Field(default_factory=list)
        int_field: int | None = None

    class ProductBlockOneForTestProvisioning(
        ProductBlockOneForTestInactive,
        AbstractProductBlockOneForTestProvisioning,
        lifecycle=[SubscriptionLifecycle.PROVISIONING],
    ):
        str_field: str | None = None
        list_field: list[int]
        int_field: int

    class ProductBlockOneForTest(
        ProductBlockOneForTestProvisioning, AbstractProductBlockOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        str_field: str
        list_field: list[int]
        int_field: int

    class AbstractProductTypeOneForTestInactive(SubscriptionModel):
        block: AbstractProductBlockOneForTestInactive

    class AbstractProductTypeOneForTestProvisioning(AbstractProductTypeOneForTestInactive):
        block: AbstractProductBlockOneForTestProvisioning

    class AbstractProductTypeOneForTest(AbstractProductTypeOneForTestProvisioning):
        block: AbstractProductBlockOneForTest

    class ProductTypeOneForTestInactive(AbstractProductTypeOneForTestInactive, is_base=True):  # noqa: F811
        block: ProductBlockOneForTestInactive

    class ProductTypeOneForTestProvisioning(  # noqa: F811
        ProductTypeOneForTestInactive,
        AbstractProductTypeOneForTestProvisioning,
        lifecycle=[SubscriptionLifecycle.PROVISIONING],
    ):
        block: ProductBlockOneForTestProvisioning

    class ProductTypeOneForTest(  # noqa: F811
        ProductTypeOneForTestProvisioning, AbstractProductTypeOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        block: ProductBlockOneForTest

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTest

    test_model = ProductTypeOneForTestInactive.from_product_id(product_id=test_product_one, customer_id=str(uuid4()))
    test_model.block = ProductBlockOneForTestInactive.new(subscription_id=test_model.subscription_id)

    product_block, product_sub_block = test_product_block_one_db

    # Check metadata
    with pytest.raises(
        ValueError, match=r"Cannot create instance of abstract class. Use one of {'ProductBlockOneForTest'}"
    ):
        AbstractProductBlockOneForTestInactive.new(subscription_id=test_model.subscription_id)
    assert AbstractProductBlockOneForTestInactive.name is None
    assert not hasattr(AbstractProductBlockOneForTestInactive, "description")
    assert not hasattr(AbstractProductBlockOneForTestInactive, "product_block_id")
    assert not hasattr(AbstractProductBlockOneForTestInactive, "tag")
    assert ProductBlockOneForTestInactive.name == "ProductBlockOneForTest"
    assert ProductBlockOneForTestInactive.description == "Test Block"
    assert ProductBlockOneForTestInactive.product_block_id == product_block.product_block_id
    assert ProductBlockOneForTestInactive.tag == "TEST"

    test_model.save()
    db.session.commit()

    test_model = AbstractProductTypeOneForTestInactive.from_subscription(test_model.subscription_id)
    assert isinstance(test_model, ProductTypeOneForTestInactive)
    assert isinstance(test_model.block, ProductBlockOneForTestInactive)

    test_model = ProductTypeOneForTestInactive.from_subscription(test_model.subscription_id)
    assert isinstance(test_model.block, ProductBlockOneForTestInactive)

    test_model.block.int_field = 1
    test_model.block.str_field = "bla"
    test_model.block.list_field = [1]

    test_model = SubscriptionModel.from_other_lifecycle(test_model, SubscriptionLifecycle.ACTIVE)
    test_model.save()
    assert isinstance(test_model.block, ProductBlockOneForTest)

    test_model = AbstractProductTypeOneForTest.from_subscription(test_model.subscription_id)
    assert isinstance(test_model.block, ProductBlockOneForTest)

    test_model = ProductTypeOneForTest.from_subscription(test_model.subscription_id)
    assert isinstance(test_model.block, ProductBlockOneForTest)

    block = AbstractProductBlockOneForTest.from_db(test_model.block.subscription_instance_id)
    assert block.model_dump() == test_model.block.model_dump()
    assert isinstance(block, ProductBlockOneForTest)

    block = ProductBlockOneForTest.from_db(test_model.block.subscription_instance_id)
    assert block.model_dump() == test_model.block.model_dump()
    assert isinstance(block, ProductBlockOneForTest)


def test_subscription_instance_list():
    def custom_conlist(t):
        return conlist(t, min_length=1, max_length=2)

    class Model(BaseModel):
        list_field: custom_conlist(int)

    with pytest.raises(ValidationError):
        Model(list_field=["a"])

    Model(list_field=[1])


def test_diff_in_db_empty(test_product_one, test_product_type_one):
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    assert ProductTypeOneForTestInactive.diff_product_in_database(test_product_one) == {}


def test_diff_in_db_when_no_fields(test_product_one, test_product_type_one):
    class Wrong(SubscriptionModel):
        pass

    assert (
        Wrong.diff_product_in_database(test_product_one)
        == {
            "TestProductOne": {
                "missing_fixed_inputs_in_model": {"test_fixed_input"},
                "missing_product_blocks_in_model": {"ProductBlockOneForTest"},
            }
        }
        != {
            "TestProductOne": {
                "missing_in_depends_on_blocks": {
                    "ProductBlockOneForTest": {"missing_product_blocks_in_db": {"SubBlockOneForTest"}}
                },
                "missing_product_blocks_in_model": {"SubBlockOneForTest"},
            }
        }
    )


def test_diff_in_db_missing_in_db(test_product_type_one):
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    product = ProductTable(
        name="TestProductEmpty", description="Test ProductTable", product_type="Test", tag="TEST", status="active"
    )

    db.session.add(product)
    db.session.commit()

    assert ProductTypeOneForTestInactive.diff_product_in_database(product.product_id) == {
        "TestProductEmpty": {
            "missing_fixed_inputs_in_db": {"test_fixed_input"},
            "missing_in_depends_on_blocks": {
                "ProductBlockOneForTest": {
                    "missing_product_blocks_in_db": {"SubBlockOneForTest"},
                    "missing_resource_types_in_db": {"int_field", "list_field", "str_field", "enum_field"},
                },
                "SubBlockOneForTest": {"missing_resource_types_in_db": {"int_field", "str_field"}},
            },
            "missing_product_blocks_in_db": {"ProductBlockOneForTest"},
        }
    }


def test_from_other_lifecycle_abstract(test_product_one):
    class AbstractProductBlockOneForTestInactive(ProductBlockModel, product_block_name="ProductBlockOneForTest"):
        str_field: str | None = None
        list_field: list[int] = Field(default_factory=list)

    class AbstractProductBlockOneForTestProvisioning(
        AbstractProductBlockOneForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
    ):
        str_field: str | None = None
        list_field: list[int]

    class AbstractProductBlockOneForTest(
        AbstractProductBlockOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        str_field: str
        list_field: list[int]

    class ProductBlockOneForTestInactive(
        AbstractProductBlockOneForTestInactive, product_block_name="ProductBlockOneForTest"
    ):
        str_field: str | None = None
        list_field: list[int] = Field(default_factory=list)
        int_field: int | None = None

    class ProductBlockOneForTestProvisioning(
        ProductBlockOneForTestInactive,
        AbstractProductBlockOneForTestProvisioning,
        lifecycle=[SubscriptionLifecycle.PROVISIONING],
    ):
        str_field: str | None = None
        list_field: list[int]
        int_field: int

    class ProductBlockOneForTest(
        ProductBlockOneForTestProvisioning, AbstractProductBlockOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        str_field: str
        list_field: list[int]
        int_field: int

    block = ProductBlockOneForTestInactive.new(subscription_id=uuid4())
    assert isinstance(block, ProductBlockOneForTestInactive)

    block.int_field = 1
    block.str_field = "bla"
    block.list_field = [1]

    active_block = ProductBlockOneForTest._from_other_lifecycle(block, SubscriptionLifecycle.ACTIVE, uuid4())

    assert isinstance(active_block, AbstractProductBlockOneForTest)
    assert isinstance(active_block, ProductBlockOneForTest)
    assert active_block.db_model == block.db_model


def test_from_other_lifecycle_sub(test_product_one, test_product_block_one, test_product_sub_block_one):
    SubBlockOneForTestInactive, _, SubBlockOneForTest = test_product_sub_block_one
    ProductBlockOneForTestInactive, _, ProductBlockOneForTest = test_product_block_one
    subscription_id = uuid4()

    block = ProductBlockOneForTestInactive.new(
        subscription_id=subscription_id, int_field=1, str_field="bla", list_field=[1], enum_field=DummyEnum.FOO
    )
    block.sub_block = SubBlockOneForTestInactive.new(subscription_id=subscription_id, int_field=1, str_field="bla")
    block.sub_block_2 = SubBlockOneForTestInactive.new(subscription_id=subscription_id, int_field=1, str_field="bla")
    block.sub_block_list = [
        SubBlockOneForTestInactive.new(subscription_id=subscription_id, int_field=1, str_field="bla")
    ]

    assert isinstance(block, ProductBlockOneForTestInactive)

    active_block = ProductBlockOneForTest._from_other_lifecycle(block, SubscriptionLifecycle.ACTIVE, uuid4())

    assert isinstance(active_block, ProductBlockOneForTest)
    assert isinstance(active_block.sub_block, SubBlockOneForTest)
    assert isinstance(active_block.sub_block_2, SubBlockOneForTest)
    assert isinstance(active_block.sub_block_list[0], SubBlockOneForTest)
    assert active_block.db_model == block.db_model
    assert active_block.sub_block.db_model == block.sub_block.db_model
    assert active_block.sub_block_2.db_model == block.sub_block_2.db_model
    assert active_block.sub_block_list[0].db_model == block.sub_block_list[0].db_model


def test_serializable_property():
    class DerivedDomainModel(DomainModel):
        @computed_field  # type: ignore[misc]
        @property
        def double_int_field(self) -> int:
            # This property is serialized
            return 2 * self.int_field

        @property
        def triple_int_field(self) -> int:
            # This property is not serialized
            return 3 * self.int_field

        int_field: int

    block = DerivedDomainModel(int_field=13)

    assert block.model_dump() == {"int_field": 13, "double_int_field": 26}


def test_inherited_serializable_property():
    class ProvisioningDomainModel(DomainModel):
        @computed_field  # type: ignore[misc]
        @property
        def double_int_field(self) -> int:
            return 2 * self.int_field

        @computed_field  # type: ignore[misc]
        @property
        def triple_int_field(self) -> int:
            return 3 * self.int_field

        int_field: int

    class ActiveDomainModel(ProvisioningDomainModel):
        @computed_field  # type: ignore[misc]
        @property
        def triple_int_field(self) -> int:
            # override the base property
            return 30 * self.int_field

    block = ActiveDomainModel(int_field=1)

    assert block.model_dump() == {"int_field": 1, "double_int_field": 2, "triple_int_field": 30}


def test_nested_serializable_property():
    """Ensure that nested serializable property's are included in the serialized model."""

    class DerivedDomainModel(DomainModel):
        @computed_field  # type: ignore[misc]
        @property
        def double_int_field(self) -> int:
            # This property is serialized
            return 2 * self.int_field

        int_field: int

    class ParentDomainModel(DomainModel):
        derived: DerivedDomainModel

    model = ParentDomainModel(derived=DerivedDomainModel(int_field=13))

    assert model.model_dump() == {"derived": {"int_field": 13, "double_int_field": 26}}


def test_property_with_tag(test_product_block_one, test_product_one, test_product_block_one_db):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    block = ProductBlockOneForTestInactive.new(int_field=1, subscription_id=uuid4())

    assert block.title == "TEST ProductBlockOneForTestInactive int_field=1"


def test_subscription_save_list_with_zero_values(
    test_product_type_one, test_product_sub_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one

    subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
    subscription.block.list_field = [10, 0, 20, 30, 40, 0, 0]
    subscription.save()
    assert subscription.block.list_field == [10, 0, 20, 30, 40, 0, 0]

    subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
    assert subscription.block.list_field == [10, 0, 20, 30, 40, 0, 0]


def test_subscription_save_bool_list_with_false_values(
    test_product_type_one, test_product_sub_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one

    subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
    subscription.block.list_field = [True, False, True, False]
    subscription.save()
    assert subscription.block.list_field == [True, False, True, False]

    subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
    assert subscription.block.list_field == [True, False, True, False]



# Copyright 2022 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from http import HTTPStatus

from test.unit_tests.config import GRAPHQL_ENDPOINT, GRAPHQL_HEADERS
from test.unit_tests.graphql.mutations.helpers import mutation_authorization


def get_start_process_mutation(
    name: str,
    payload: dict,
) -> bytes:
    query = """
mutation StartProcessMutation ($name: String!, $payload: Payload!) {
    startProcess(name: $name, payload: $payload) {
        ... on ProcessCreated {
            id
        }
        ... on MutationError {
            message
            details
        }
    }
}
    """
    return json.dumps(
        {
            "operationName": "StartProcessMutation",
            "query": query,
            "variables": {
                "name": name,
                "payload": payload,
            },
        }
    ).encode("utf-8")


def test_process_not_found(httpx_mock, test_client):
    # given

    unknown_wf = "unknown_wf"

    # when

    query = get_start_process_mutation(name=unknown_wf, payload={"payload": {}})

    with mutation_authorization():
        response = test_client.post(GRAPHQL_ENDPOINT, content=query, headers=GRAPHQL_HEADERS)

    # then

    expected = {
        "data": {"startProcess": {"details": "404: Workflow does not exist", "message": "Could not create process"}}
    }

    assert response.status_code == HTTPStatus.OK
    assert response.json() == expected


def test_process_started(httpx_mock, test_client, generic_product_type_1):
    # given

    known_wf = "task_validate_products"

    # when

    query = get_start_process_mutation(name=known_wf, payload={"payload": [{}]})

    with mutation_authorization():
        response = test_client.post(GRAPHQL_ENDPOINT, content=query, headers=GRAPHQL_HEADERS)

    # then
    assert response.status_code == HTTPStatus.OK
    data = response.json()
    assert data.get("data", {}).get("startProcess", {}).get("id")



import structlog
from orchestrator.forms import FormPage
from orchestrator.forms.validators import DisplaySubscription
from orchestrator.types import InputForm, State, UUIDstr
from orchestrator.workflow import StepList, begin, step
from orchestrator.workflows.utils import terminate_workflow

from products.product_types.example2 import Example2

logger = structlog.get_logger(__name__)


def terminate_initial_input_form_generator(subscription_id: UUIDstr, customer_id: UUIDstr) -> InputForm:
    temp_subscription_id = subscription_id

    class TerminateExample2Form(FormPage):
        subscription_id: DisplaySubscription = temp_subscription_id  # type: ignore

    return TerminateExample2Form


@step("Delete subscription from OSS/BSS")
def delete_subscription_from_oss_bss(subscription: Example2) -> State:
    # TODO: add actual call to OSS/BSS to delete subscription

    return {}


additional_steps = begin


@terminate_workflow(
    "Terminate example2", initial_input_form=terminate_initial_input_form_generator, additional_steps=additional_steps
)
def terminate_example2() -> StepList:
    return (
        begin >> delete_subscription_from_oss_bss
        # TODO: fill in additional steps if needed
    )
